import numpy as np
from math import floor
from miplib.data.containers.image import Image
from miplib.data.indexers import polar_indexers as indexers


def fft_filter(image, threshold, kind='low'):
    assert isinstance(image, Image)

    spacing = image.spacing

    fft_image = np.fft.fftshift(np.fft.fftn(image))

    if kind == 'low':
        indexer = indexers.PolarLowPassIndexer(image.shape)
    elif kind == 'high':
        indexer = indexers.PolarHighPassIndexer(image.shape)
    else:
        raise ValueError("Unknown filter kind: {}".format(kind))

    r_max = floor(min(image.shape) / 2)

    fft_image *= indexer[threshold*r_max]

    return Image(np.abs(np.fft.ifftn(fft_image).real), spacing)


def butterworth_fft_filter(image, threshold, n=3):
    """Create low-pass 2D Butterworth filter.
    :Parameters:
       size : tuple
           size of the filter
       cutoff : float
           relative cutoff frequency of the filter (0 - 1.0)
       n : int, optional
           order of the filter, the higher n is the sharper
           the transition is.
    :Returns:
       numpy.ndarray
         filter kernel in 2D centered
   """
    if not 0 < threshold <= 1.0:
        raise ValueError('Cutoff frequency must be between 0 and 1.0')

    if not isinstance(n, int):
        raise ValueError('n must be an integer >= 1')

    assert isinstance(image, Image)

    spacing = image.spacing

    # Create Fourier grid
    r = indexers.SimplePolarIndexer(image.shape).r

    threshold *= image.shape[0]

    butter = 1.0 / (1.0 + (r / threshold) ** (2 * n))  # The filter

    fft_image = np.fft.fftshift(np.fft.fftn(image))
    fft_image *= butter

    return Image(np.abs(np.fft.ifftn(fft_image).real), spacing)


def gaussian_fft_filter(image, threshold):
    """
    Create low-pass 2D Gaussian filter.
    :Parameters:
       size : tuple
           size of the filter
       cutoff : float
           relative cutoff frequency of the filter (0 - 1.0)
       n : int, optional
           order of the filter, the higher n is the sharper
           the transition is.
    :Returns:
       numpy.ndarray:  filter kernel in 2D centered
   """
    if not 0 < threshold <= 1.0:
        raise ValueError('Cutoff frequency must be between 0 and 1.0')

    assert isinstance(image, Image)

    spacing = image.spacing

    # Create Fourier grid
    r = indexers.SimplePolarIndexer(image.shape).r

    r /= image.shape[0]

    gauss = np.exp(-(r**2/(2*(threshold**2))))
    fft_image = np.fft.fftshift(np.fft.fftn(image))
    fft_image *= gauss

    return Image(np.abs(np.fft.ifftn(fft_image).real), spacing)




"""
Sami Koho - IIT

This file contains classes for generating different kinds of complex
indexing structures (masks).
"""

import numpy as np


class SimplePolarIndexer(object):
    """
    Basic indexer for a polar/spherical coordinate system
    """
    def __init__(self, shape):
        assert isinstance(shape, tuple) or \
               isinstance(shape, list) or \
               isinstance(shape, np.ndarray)
        assert 1 < len(shape) < 4

        # Create Fourier grid
        axes = (np.arange(-np.floor(i / 2.0), np.ceil(i / 2.0)) for i in shape)

        meshgrid = np.meshgrid(*axes)
        self.r = np.sqrt(sum([axis**2 for axis in meshgrid]))

    def __getitem__(self, item):
        return self.r == item


class PolarLowPassIndexer(SimplePolarIndexer):
    """
    Generates a low-pass mask in the polar coordinate system, i.e. points
    closer than the specified distance will be selected.
    """
    def __getitem__(self, item):
        return self.r < item


class PolarHighPassIndexer(SimplePolarIndexer):
    """
    Generates a high-pass mask in the polar coordinate system, i.e. points
    farther than the specified distance will be selected.
    """
    def __getitem__(self, item):
        return self.r > item






import numpy as np
from scipy import ndimage


from miplib.data.containers.image import Image


def make_local_intensity_based_mask(image, threshold, kernel_size=100, invert=False):
    assert isinstance(image, Image)

    blurred_image = ndimage.uniform_filter(image, size=kernel_size)

    peaks = np.percentile(blurred_image, threshold)
    mask = np.where(blurred_image >= peaks, 1, 0)
    if invert:
        return np.invert(mask.astype(bool))
    else:
        return mask



"""
File:        image.py
Author:      Sami Koho (sami.koho@gmail.com)

Description:
This file contains a simple class for storing image data.
"""

import argparse

import numpy


class Image(numpy.ndarray):
    """
    A very simple extension to numpy.ndarray, to contain image data and
    metadata.
    """

    # region Initialization

    def __new__(cls, images, spacing, filename=None):
        obj = numpy.asarray(images).view(cls)

        obj.spacing = list(spacing)
        obj.filename = filename

        return obj

    def __array__finalize__(self, obj):

        self.spacing = getattr(obj, 'spacing')
        self.filename = getattr(obj, 'filename', None)
    # endregion

    # region Properties

    # @property
    # def spacing(self): return self._spacing
    #
    # @spacing.setter
    # def spacing(self, value):
    #     if len(value) != len(self.shape):
    #         raise ValueError("You should define spacing for every dimension")
    #     else:
    #         self._spacing = value

    # endregion

# region Command Line Arguments (refactor)
def get_options(parser):
    """
    Command-line options for the image I/O
    """
    assert isinstance(parser, argparse.ArgumentParser)
    group = parser.add_argument_group("Image I/O", "Options for image file I/O")
    # Parameters for controlling how image files are handled
    group.add_argument(
        "--imagej",
        help="Defines wheter the image are in ImageJ tiff format, "
             "and thus contain the pixel size info etc in the TIFF tags. "
             "By default true",
        action="store_true"
    )
    group.add_argument(
        "--rgb-channel",
        help="Select which channel in an RGB image is to be used for quality"
             " analysis",
        dest="rgb_channel",
        type=int,
        choices=[0, 1, 2],
        default=1
    )
     # File filtering for batch mode processing
    parser.add_argument(
        "--average-filter",
        dest="average_filter",
        type=int,
        default=0,
        help="Analyze only images with similar amount of detail, by selecting a "
             "grayscale average pixel value threshold here"
    )
    parser.add_argument(
        "--file-filter",
        dest="file_filter",
        default=None,
        help="Define a common string in the files to be analysed"
    )
    return parser
# endregion














import numpy as np
from scipy.ndimage import interpolation

from . import ndarray
from miplib.data.containers.image import Image


def zoom_to_isotropic_spacing(image, order=3):
    """
    Resize an Image to isotropic pixel spacing.

    :param image:   a Image object
    :param order:   the spline interpolation type
    :return:        a isotropically spaced Image
    """
    assert isinstance(image, Image)

    spacing = image.spacing
    old_shape = image.shape
    min_spacing = min(spacing)
    zoom = tuple(pixel_spacing / min_spacing for pixel_spacing in spacing)
    new_shape = tuple(int(pixels * dim_zoom) for (pixels, dim_zoom) in zip(old_shape, zoom))

    if new_shape == old_shape:
        return image
    else:
        return resize(image, new_shape, order)

def zoom_to_spacing(image, spacing, order=3, verbose=False):

    assert isinstance(image, Image)
    assert image.ndim == len(spacing)

    zoom = tuple(i/j for i, j in zip(image.spacing, spacing))
    if verbose:
        print("The zoom is ", zoom)

    array = interpolation.zoom(image, zoom, order=order)

    return Image(array, spacing)


def resize(image, size, order=3, verbose=False):  # type: (Image, tuple) -> Image
    """
    Resize the image, using interpolation.

    :param order:   The interpolation type defined as order of the b-spline
    :param image:   The MyImage object.
    :param size:    A tuple of new image dimensions.

    """
    assert isinstance(size, tuple)
    assert isinstance(image, Image)

    zoom = [float(a) / b for a, b in zip(size, image.shape)]
    if verbose:
        print("The zoom is %s" % zoom)

    array = interpolation.zoom(image, tuple(zoom), order=order)
    spacing = tuple(i / j for i, j in zip(image.spacing, zoom))

    return Image(array, spacing)


def apply_hanning(image):  # type: (Image) -> Image
    """
    Apply Hanning window to the image.

    :return:
    """

    windows = (np.hanning(i) for i in image.shape)

    result = Image(image.astype('float64'), image.spacing)
    for window in windows:
        result *= window

    return result


def zero_pad_to_shape(image, shape):
    """
    Apply zero padding to cast an Image into the given shape. The zero padding
    will be applied evenly on all sides of the image.

    :param image: an Image object
    :param shape: a shape tuple
    :return:      the zero padded Image
    """
    assert isinstance(image, Image)

    if image.shape == shape:
        return image
    else:
        return Image(ndarray.expand_to_shape(image, shape), image.spacing)


def zero_pad_to_matching_shape(image1, image2):
    """
    Apply zero padding to make the size of two Images match.
    :param image1: an Image object
    :param image2: an Image object
    :return:       zero padded image1 and image2
    """

    assert isinstance(image1, Image)
    assert isinstance(image2, Image)

    shape = tuple(max(x, y) for x, y in zip(image1.shape, image2.shape))

    if any(map(lambda x,y: x != y, image1.shape, shape)):
        image1 = zero_pad_to_shape(image1, shape)
    if any(map(lambda x,y: x != y, image2.shape, shape)):
        image2 = zero_pad_to_shape(image2, shape)

    return image1, image2


def remove_zero_padding(image, shape):
    """

    :param image: The zero padded image
    :param shape: The original image size (before padding)
    :return:
    """

    assert isinstance(image, Image)
    assert len(shape) == image.ndim

    return Image(ndarray.contract_to_shape(image, shape), image.spacing)


def checkerboard_split(image, disable_3d_sum = False):
    """
    Splits an image in two, by using a checkerboard pattern.

    :param image:   a miplib Image
    :return:        two miplib Images
    """
    assert isinstance(image, Image)

    # Make an index chess board structure
    shape = image.shape
    odd_index = list(np.arange(1, shape[i], 2) for i in range(len(shape)))
    even_index = list(np.arange(0, shape[i], 2) for i in range(len(shape)))

    # Create the two pseudo images
    if image.ndim == 2:
        image1 = image[odd_index[0], :][:, odd_index[1]]
        image2 = image[even_index[0], :][:, even_index[1]]
    else:
        if disable_3d_sum:
            image1 = image[odd_index[0], :, :][:, odd_index[1], :][:, :, odd_index[2]]
            image2 = image[even_index[0], :, :][:, even_index[1], :][:, :, even_index[2]]

        else:
            image1 = image.astype(np.uint32)[even_index[0], :, :][:, odd_index[1], :][:, :, odd_index[2]] + \
                     image.astype(np.uint32)[odd_index[0], :, :][:, odd_index[1], :][:, :, odd_index[2]]

            image2 = image.astype(np.uint32)[even_index[0], :, :][:, even_index[1], :][:, :, even_index[2]] + \
                     image.astype(np.uint32)[odd_index[0], :, :][:, even_index[1], :][:, :, even_index[2]]

    # image1.spacing = tuple(i * np.sqrt(2) for i in image.spacing)
    image1.spacing = image.spacing
    image2.spacing = image1.spacing

    return image1, image2


def reverse_checkerboard_split(image, disable_3d_sum = False):
    """
    Splits an image in two, by using a checkerboard pattern.

    :param image:   a miplib Image
    :return:        two miplib Images
    """
    assert isinstance(image, Image)

    # Make an index chess board structure
    shape = image.shape
    odd_index = list(np.arange(1, shape[i], 2) for i in range(len(shape)))
    even_index = list(np.arange(0, shape[i], 2) for i in range(len(shape)))

    # Create the two pseudo images
    if image.ndim == 2:
        image1 = image[odd_index[0], :][:, even_index[1]]
        image2 = image[even_index[0], :][:, odd_index[1]]
    else:
        if disable_3d_sum:
            image1 = image[odd_index[0], :, :][:, odd_index[1], :][:, :, even_index[2]]
            image2 = image[even_index[0], :, :][:, even_index[1], :][:, :, odd_index[2]]

        else:
            image1 = image.astype(np.uint32)[even_index[0], :, :][:, odd_index[1], :][:, :, even_index[2]] + \
                     image.astype(np.uint32)[odd_index[0], :, :][:, even_index[1], :][:, :, odd_index[2]]

            image2 = image.astype(np.uint32)[even_index[0], :, :][:, even_index[1], :][:, :, odd_index[2]] + \
                     image.astype(np.uint32)[odd_index[0], :, :][:, odd_index[1], :][:, :, even_index[2]]

    #image1.spacing = tuple(i * np.sqrt(2) for i in image.spacing)
    image1.spacing = image.spacing
    image2.spacing = image1.spacing

    return image1, image2

def summed_checkerboard_split(image):
    """
    Splits an image in two, by using a checkerboard pattern and diagonal pixels
    in each 4 pixel group (2D) case and orthogonal diagonal groups (never adjacent)
    in 3D case.

    :param image:   a miplib Image
    :return:        two miplib Images
    """
    assert isinstance(image, Image)

    # Make an index chess board structure
    shape = image.shape
    odd_index = list(np.arange(1, shape[i], 2) for i in range(len(shape)))
    even_index = list(np.arange(0, shape[i], 2) for i in range(len(shape)))

    # Create the two pseudo images
    if image.ndim == 2:
        image1 = image[odd_index[0], :][:, odd_index[1]] + image[even_index[0], :][:, even_index[1]]
        image2 = image[odd_index[0], :][:, even_index[1]] + image[even_index[0], :][:, odd_index[1]]

        image1.spacing = tuple(i * 2 for i in image.spacing)
        image2.spacing = image1.spacing
    else:
        image1 = image.astype(np.uint32)[even_index[0], :, :][:, odd_index[1], :][:, :, odd_index[2]] + \
                 image.astype(np.uint32)[odd_index[0], :, :][:, odd_index[1], :][:, :, odd_index[2]] + \
                 image.astype(np.uint32)[even_index[0], :, :][:, even_index[1], :][:, :, even_index[2]] + \
                 image.astype(np.uint32)[odd_index[0], :, :][:, even_index[1], :][:, :, even_index[2]]

        image2 = image.astype(np.uint32)[even_index[0], :, :][:, odd_index[1], :][:, :, even_index[2]] + \
                 image.astype(np.uint32)[odd_index[0], :, :][:, odd_index[1], :][:, :, even_index[2]] + \
                 image.astype(np.uint32)[even_index[0], :, :][:, even_index[1], :][:, :, odd_index[2]] +\
                 image.astype(np.uint32)[odd_index[0], :, :][:, even_index[1], :][:, :, odd_index[2]]

        image1.spacing = tuple(i * 2 for i in image.spacing)
        image2.spacing = image1.spacing


    return image1, image2


def zero_pad_to_cube(image):
    """
    Apply zero padding to cast an image into a cube shape (to match the number
    of pixels in all dimensions)
    :param image: an Image object
    :return:      zero padded input image, or the original, if already a cube
    """
    assert isinstance(image, Image)

    original_shape = image.shape
    nmax = max(original_shape)
    square_shape = (nmax,) * image.ndim
    if square_shape != original_shape:
        return zero_pad_to_shape(image, square_shape)
    else:
        return image


def crop_to_largest_square(image, physical_dims=False):
    """
    Crops an image into a largest square shape that fits inside the image area in all
    dimensions. The cropping can bone either in physical units or in pixels (typically pixels)
    :param image: an image Object
    :return: the cropped image
    """
    assert isinstance(image, Image)

    if physical_dims:
        shape_real = list(x*y for x, y in zip(image.shape, image.spacing))
        min_shape_real = (min(*shape_real), ) * image.ndim
        min_shape_px = list(x / y for x, y in zip(min_shape_real, image.spacing))
    else:
        min_shape_px = (min(*image.shape),) * image.ndim

    return remove_zero_padding(image, min_shape_px)


def crop_to_shape(image, shape, offset):
    """
    Crop image to shape.

    :param image:   An N-dimensional Image to be cropped
    :type image:    Image
    :param shape:   The new, cropped size; should be greater or equal than
                    the original
    :type shape:    tuple
    :return:        Returns the cropped image as an Image object
    """
    assert isinstance(image, Image)
    assert image.ndim == len(shape) == len(offset)
    assert all((v + x <= y for v, x, y in zip(offset, shape, image.shape)))

    crop_idx = tuple(slice(start, size + start) for start, size in zip(offset, shape))

    return Image(image[crop_idx], image.spacing)


def noisy(image, noise_type):
    """
    Parameters
    ----------
    image :
        Input image data. Will be converted to float.
    noise_type : str
        One of the following strings, selecting the type of noise to add:

        'gauss'     Gaussian-distributed additive noise.
        'poisson'   Poisson-distributed noise generated from the data.
        's&p'       Replaces random pixels with 0  or 1.
        'speckle'   Multiplicative noise using out = image + n*image,where
                    n is uniform noise with specified mean & variance.
    """
    assert isinstance(image, Image)
    assert image.ndim < 4
    spacing = image.spacing

    if noise_type == "gauss":
        mean = 0
        var = 0.1
        sigma = var ** 0.5
        gauss = np.random.normal(mean, sigma, image.shape)
        gauss = gauss.reshape(image.shape)
        return Image(image + gauss, spacing)
    elif noise_type == "s&p":
        s_vs_p = 0.5
        amount = 0.004
        out = np.copy(image)
        # Salt mode
        num_salt = np.ceil(amount * image.size * s_vs_p)
        coords = [np.random.randint(0, i - 1, int(num_salt))
                  for i in image.shape]
        out[coords] = 1

        # Pepper mode
        num_pepper = np.ceil(amount * image.size * (1. - s_vs_p))
        coords = [np.random.randint(0, i - 1, int(num_pepper))
                  for i in image.shape]
        out[coords] = 0
        return Image(out, spacing)
    elif noise_type == "poisson":
        vals = 2 ** np.ceil(np.log2(len(np.unique(image))))
        return Image(np.random.poisson(image * vals) / float(vals), spacing)
    elif noise_type == "speckle":
        gauss = np.random.standard_normal(image.shape).reshape(image.shape)
        return Image(image + image * gauss, spacing)


def enhance_contrast(image, percent_saturated=0.3, out_type=np.uint8):
    """
    Performs historgram stretching (not equalization), with a given percentage
    :param percent_saturated of pixels saturated in the output.

    :param image: an Image object
    :param percent_saturated: Percentage value of saturated pixels. Defaults to 0.3
    :param out_type: The type of the output image. The default is 8-bit uint
    :return: an Image with intensity values rescaled to the whole dynamic range
    """

    assert isinstance(image, Image)

    percent_saturated /= 100

    spacing = image.spacing

    if out_type == np.uint8:
        out_max = 255
        out_min = 0
    else:
        raise ValueError("Not supported output type {}".format(out_type))

    # Get Input Image Min/Max from histogram
    histogram, bin_edges = np.histogram(image, bins=250, density=True)
    cumulative = np.cumsum(histogram * np.diff(bin_edges))

    in_max = bin_edges[1:][cumulative >= 1.0 - percent_saturated].min()

    to_zero = cumulative <= percent_saturated
    if not np.any(to_zero):
        in_min = image.min()
    else:
        in_min = bin_edges[1:][to_zero].max()

    # Trim and rescale
    image = np.clip(image, in_min, in_max)
    image *= (out_max-out_min)/image.max()
    return Image(image.astype(out_type), spacing)


def rescale_to_8_bit(image):
    """
    Converts an Image into 8-bit (typically for saving)
    :param image: an Image object
    :return: a 8-bit version of the Image
    """
    assert isinstance(image, Image)
    return Image((image*(255.0/image.max())).astype(np.uint8), image.spacing)




def flip_image(image):

    assert isinstance(image, Image)

    indexer = (np.s_[::-1],) * image.ndim

    return Image(image[indexer], image.spacing)


def translate_image(image, shift):
    """
    Apply a circular shift to an image

    :param image: An Image object
    :param shift: The shift as a single numeric value
    :return: returns the translated image.
    """
    fft_image = np.fft.fftshift(np.fft.fft2(image))

    shape = fft_image.shape
    axes = (np.arange(-np.floor(i / 2.0), np.ceil(i / 2.0)) for i in shape)
    axes = (i / (2 * i.max()) for i in axes)
    y, x = np.meshgrid(*axes)

    xx = np.zeros(fft_image.shape, dtype=np.complex64)
    xx.real[:] = np.cos(2 * np.pi * shift * x)
    xx.imag[:] = np.sin(-2 * np.pi * shift * x)

    yy = np.zeros(fft_image.shape, dtype=np.complex64)
    yy.real[:] = np.cos(2 * np.pi * shift * y)
    yy.imag[:] = np.sin(-2 * np.pi * shift * y)

    multiplier = xx * yy

    result = np.abs(np.fft.ifftn(fft_image * multiplier).real)

    return Image(result, image.spacing)

def maximum_projection(image, axis=0):
    """ Generate a maximum projection image along an axis
    
    :param image: an image
    :type image: Image
    :param axis: the axis on which the projeciton is to be calculated, defaults to 0
    :type axis: int, optional
    :return: a maximum projection image, with one dimension less thatn the input image
    :rtype: Image
    """
    assert isinstance(image, Image)
    spacing = (image.spacing[s] for s in filter(lambda x : x != axis, range(image.ndim)))
    return  Image(np.amax(image, axis=axis), spacing)



"""
Options for multi-view image fusion
"""
import argparse
from miplib.ui.cli.argparse_helpers import parse_range_list

def get_fusion_options_group(parser):
    assert isinstance(parser, argparse.ArgumentParser)
    group = parser.add_argument_group("Fusion", "Options for image fusion")

    group.add_argument(
        '--disable-cuda',
        action='store_true'
    )
    group.add_argument(
        '--max-nof-iterations',
        type=int,
        default=100,
        help='Specify maximum number of iterations.'
    )
    group.add_argument(
        '--convergence-epsilon',
        dest='convergence_epsilon',
        type=float,
        default=0.05,
        help='Specify small positive number that determines '
             'the window for convergence criteria.'
    )

    group.add_argument(
        '--first-estimate',
        choices=['first_image',
                 'first_image_mean',
                 'sum_of_originals',
                 'sum_of_registered',
                 'average_of_all',
                 'simple_fusion',
                 'constant'],
        default='first_image_mean',
        help='Specify first estimate for iteration.'
    )

    group.add_argument(
        '--estimate-constant',
        dest='estimate_constant',
        type=float,
        default=1.0
    )

    group.add_argument(
        '--wiener-snr',
        type=float,
        default=100.0
    )

    group.add_argument(
        '--save-intermediate-results',
        action='store_true',
        help='Save intermediate results.'
    )

    group.add_argument(
        '--output-cast',
        dest='output_cast',
        action='store_true',
        help='By default the fusion output is returned as a 32-bit image'
             'This switch can be used to enable 8-bit unsigned output'
    )

    group.add_argument(
        '--fusion-method',
        dest='fusion_method',
        choices=['multiplicative', 'multiplicative-opt', 'summative',
                 'summative-opt'],
        default='summative'
    )

    group.add_argument(
        '--blocks',
        dest='num_blocks',
        type=int,
        default=1,
        help="Define the number of blocks you want to break the images into"
             "for the image fusion. This argument defaults to 1, which means"
             "that the entire image will be used -- you should define a larger"
             "number to optimize memory consumption"
    )
    group.add_argument(
        '--rltv-stop-tau',
        type=float,
        default=0.002,
        help='Specify parameter for tau-stopping criteria.'
    )
    group.add_argument(
        '--rltv-lambda',
        type=float,
        default=0,
        help="Enable Total Variation regularization by selecting value > 0"
    )

    group.add_argument(
        '--pad',
        dest='block_pad',
        type=int,
        default=0,
        help='The amount of padding to apply to a fusion block.'
    )
    group.add_argument(
        '--fuse-views',
        dest='fuse_views',
        type=parse_range_list,
        default=-1
    )
    group.add_argument(
        '--memmap-estimates',
        action='store_true'
    )

    group.add_argument(
        '--disable-tau1',
        action='store_true'
    )

    group.add_argument(
        '--disable-fft-psf-memmap',
        action='store_true'
    )
    return parser




"""
Various utilities that are used to convert command line parameters into
data types that the progrma understands.
"""
import os

file_extensions = ['.tif', '.lsm', 'tiff', '.raw', '.data']


def get_user_input(message):
    """
    A method to ask question. The answer needs to be yes or no.

    Parameters
    ----------
    :param message  string, the question

    Returns
    -------

    Return a boolean: True for Yes, False for No
    """
    while True:
        answer = input(message)
        if answer in ('y', 'Y', 'yes', 'YES'):
            return True
        elif answer in ('n', 'N', 'no', 'No'):
            return False
        else:
            print("Unkown command. Please state yes or no")


def get_path_dir(path, suffix):
    """ Return a directory name with suffix that will be used to save data
    related to given path.
    """
    if os.path.isfile(path):
        path_dir = path + '.' + suffix
    elif os.path.isdir(path):
        path_dir = os.path.join(path, suffix)
    elif os.path.exists(path):
        raise ValueError('Not a file or directory: %r' % path)
    else:
        base, ext = os.path.splitext(path)
        if ext in file_extensions:
            path_dir = path + '.' + suffix
        else:
            path_dir = os.path.join(path, suffix)
    return path_dir


def get_full_path(path, prefix):
    """
    :param path:    Path to a file (string)
    :param prefix:  Path prefix, if applicable. Used in cases in
                    which the path argument is not an absolute
                    path
    :return:        Returns the absolute path, if the file is found,
                    None otherwise
    """
    if not os.path.isfile(path):
        path = os.path.join(prefix, path)
        if not os.path.isfile(path):
            raise ValueError('Not a valid file %s' % path)
    return path


def get_filename_and_extension(path):
    """
    Returns a filename and the file extension. The filename cna
    be either a simlpe filename or a full path.

    :param path:
    :return:
    """
    filename = path.split('.')[:-1]
    extension = path.split('.')[-1]

    return filename, extension



from ..containers.array_detector_data import ArrayDetectorData
from ..containers.image_data import ImageData


def convert_to_imagedata(data, path, data_type="original"):
    """
    A very simple converter to save ArrayDetectorData into the HDF5 data structure
    used in MIPLIB for large image datasets.

    :param data: an ArrayDetectorDAta object
    :param path: full path to the new hdf5 file. Should end with .hdf5
    :return: returns a handle to the new hdf5 file. The file is left open, so remember
    to call close() method in order to ensure that all the data is written on the disk.
    """
    assert isinstance(data, ArrayDetectorData)

    image_data = ImageData(path)

    for gate_idx in range(data.ngates):
        for det_idx in range(data.ndetectors):
            temp = data[gate_idx, det_idx]
            if data_type == "original":
                image_data.add_original_image(temp, 100, det_idx, gate_idx, 0, temp.spacing)
            elif data_type == "registered":
                image_data.add_registered_image(temp, 100, det_idx, gate_idx, 0, temp.spacing)
            elif data_type == "psf":
                image_data.add_psf(temp, 100, det_idx, gate_idx, 0, temp.spacing)

    return image_data


from unittest import TestCase

from ..dictionary import FixedDictionary


class TestFixedDictionary(TestCase):
    def test_set_get_item(self):
        dictionary = FixedDictionary(("key1", "key2", "key3"))

        dictionary["key1"] = 23

        self.assertEqual(dictionary["key1"], 23)

    def test_set_wrong_item(self):
        dictionary = FixedDictionary(("key1", "key2", "key3"))
        with self.assertRaises(KeyError):
            dictionary["key5"] = 25

    def test_contents(self):
        dictionary = FixedDictionary(("key1", "key2", "key3"))

        dictionary["key2"] = 23
        dictionary["key1"] = "temp"
        dictionary["key3"] = (1, 2, 3)

        keys, values = dictionary.contents

        self.assertListEqual(keys, ['key3', 'key2', 'key1'])
        self.assertListEqual(values, [(1, 2, 3), 23, 'temp'])








import os
import argparse

from itertools import chain

def parse_range_list(rngs):
    """ This parser type was created to enable the input of numeric ranges, 
    such as "2, 5, 7-11, 26". It returns a sorted list of integers.
    
    Arguments:
        rngs {string} -- A string containing comma delimited list of ranges, 
        e.g. "2, 5, 7-11, 26". A range should consist of a start and end (3-6)
        or a single integer number. 
    
    Raises:
        ValueError: Raised if a bad range, e.g. 7-11-3 is given. 
    
    Returns:
        tuple -- A tuple with the selected indices. The above range for example,
        "2, 5, 7-11, 26"  will generate a tuple(2, 5, 7, 8, 9, 10, 11, 26)
    """
    def parse_range(rng):
        parts = rng.split('-')
        if 1 > len(parts) > 2:
            raise ValueError("Bad range: '%s'" % (rng,))
        parts = [int(i) for i in parts]
        start = parts[0]
        end = start if len(parts) == 1 else parts[1]
        if start > end:
            end, start = start, end
        return list(range(start, end + 1))

    return sorted(set(chain(*[parse_range(rng) for rng in rngs.split(',')])))


def parseFromToString(string):
    return list(int(i) for i in string.split("to"))


def ensure_positive(number):
    """ Check that a positive number is inserted
    
    Arguments:
        number {string} -- a string of a number, may be float or an int
    
    Raises:
        argparse.ArgumentTypeError: raises an error if argument is not a number
        or if the number is negative
    
    Returns:
        float -- returns the number as as a float
    """
    try:
        number = float(number)
    except ValueError:
        msg = "You must enter a number"
        raise argparse.ArgumentTypeError(msg)
    if number <= 0:
        raise argparse.ArgumentTypeError("The value should be greater than zero")

    return number

    import os
import argparse

def parse_int_tuple(string):
    """ Converts a string of comma separated integer digits into a tuple of ints

    Arguments:
    string {string} -- The input string

    Returns:
    tuple -- A tuple of integers
    """
    
    return tuple(int(i) for i in string.split(','))


def parse_float_tuple(string):
    """ Converts a string of comma separated floating point numbers 
    (. for decimal) into a tuple of floating point numbers.
    
    Arguments:
        string {string} -- The input string
    
    Returns:
        tuple -- The tuple of floats.
    """
    return tuple(float(i) for i in string.split(','))

def parse_is_dir(dirname):
    """ Checks if a path is an actual directory

    
    Arguments:
        dirname {string} -- A path to a directory
    
    Raises:
        argparse.ArgumentTypeError: Raises a parse error if the directory does
        not exist.
    
    Returns:
        string -- Returns the directory, if it exists.
    """
    if not os.path.isdir(dirname):
        msg = "{0} is not a directory".format(dirname)
        raise argparse.ArgumentTypeError(msg)
    else:
        return dirname



from miplib.data.core.dictionary import FixedDictionary

import pandas as pd
import numpy as np


class FourierCorrelationDataCollection(object):
    """
    A container for the directional Fourier correlation data
    """
    def __init__(self):
        self._data = dict()

        self.iter_index = 0

    def __setitem__(self, key, value):
        assert isinstance(key, (int, np.integer))
        assert isinstance(value, FourierCorrelationData)

        self._data[str(key)] = value

    def __getitem__(self, key):
        return self._data[str(key)]

    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = list(self._data.items())[self.iter_index]
        except IndexError:
            self.iter_index = 0
            raise StopIteration

        self.iter_index += 1
        return item

    def __len__(self):
        return len(self._data)

    def clear(self):
        self._data.clear()

    def items(self):
        return list(self._data.items())

    def nitems(self):
        return len(self._data)

    def as_dataframe(self, include_results=False):
        """
        Convert a FourierCorrelationDataCollection object into a Pandas
        dataframe. Only returns the raw Fourier correlation data,
        not the processed results.

        :return: A dataframe with columns: Angle (categorical), Correlation (Y),
                 Frequency (X) and nPoints (number of points in each bin)
        """
        df = pd.DataFrame(columns=['Correlation', 'Frequency', 'nPoints', 'Angle'])

        for key, dataset in self._data.items():
            df_temp = dataset.as_dataframe(include_results=include_results)

            angle = np.full(len(df_temp), int(key), dtype=np.int64)
            df_temp['Angle'] = angle

            df = pd.concat([df, df_temp], ignore_index=True)

        df['Angle'] = df['Angle'].astype('category')
        return df


class FourierCorrelationData(object):
    """
    A datatype for FRC data

    """
    #todo: the dictionary format here is a bit clumsy. Maybe change to a simpler structure

    def __init__(self, data=None):

        correlation_keys = "correlation frequency points-x-bin curve-fit " \
                           "curve-fit-coefficients"
        resolution_keys = "threshold criterion resolution-point " \
                          "resolution-threshold-coefficients resolution spacing"

        self.resolution = FixedDictionary(resolution_keys.split())
        self.correlation = FixedDictionary(correlation_keys.split())

        if data is not None:
            assert isinstance(data, dict)

            for key, value in data.items():
                if key in self.resolution.keys:
                    self.resolution[key] = value
                elif key in self.correlation.keys:
                    self.correlation[key] = value
                else:
                    raise ValueError("Unknown key found in the initialization data")

    def as_dataframe(self, include_results=False):
        """
        Convert a FourierCorrelationData object into a Pandas
        dataframe. Only returns the raw Fourier correlation data,
        not the processed results.

        :return: A dataframe with columns: Correlation (Y), Frequency (X) and
                 nPoints (number of points in each bin)
        """
        if include_results is False:
            to_df = {
                'Correlation': self.correlation["correlation"],
                'Frequency': self.correlation["frequency"],
                'nPoints': self.correlation["points-x-bin"],
            }
        else:
            resolution = np.full(self.correlation["correlation"].shape,
                                 self.resolution["resolution"],
                                 dtype=np.float32)
            resolution_point_x = np.full(self.correlation["correlation"].shape,
                                         self.resolution["resolution-point"][0],
                                         dtype=np.float32)
            resolution_point_y = np.full(self.correlation["correlation"].shape,
                                         self.resolution["resolution-point"][1],
                                         dtype=np.float32)
            threshold = self.resolution["threshold"],

            to_df = {
                'Correlation': self.correlation["correlation"],
                'Frequency': self.correlation["frequency"],
                'nPoints': self.correlation["points-x-bin"],
                'Resolution': resolution,
                'Resolution_X': resolution_point_x,
                'Resolution_Y': resolution_point_y,
                'Threshold': threshold

            }

        return pd.DataFrame(to_df)



import numpy as np
from scipy.io import loadmat

from miplib.data.containers.array_detector_data import ArrayDetectorData
from miplib.data.containers.image import Image


def read_carma_mat(filename):
    """
    A simple implementation for the carma file import in Python
    :param filename: Path to the Carma .mat file
    :return: Returns a 2D nested list of miplib Image objects. The first dimension
             of the list corresponds to the Photosensors count and second, to the
             Detectors count
    """
    assert filename.endswith(".mat")
    #measurement = "meas_" + filename.split('/')[-1].split('.')[0]
    data = loadmat(filename)

    # Find measurement name (in case someone renamed the file)
    for key in list(data.keys()):
        if 'meas_' in key:
            data = data[key]
            break

    # Get necessary metadata
    spacing = list(data['PixelSize'][0][0][0][::-1])
    shape = list(data['Size'][0][0][0][::-1])

    detectors = int(data['DetectorsCount'][0][0][0])
    photosensors = len(data["PhotosensorsTime"][0][0][0])

    # Initialize data container
    container = ArrayDetectorData(detectors, photosensors)

    # Read images
    for i in range(photosensors):
        for j in range(detectors):
            name = 'pixel_d{}_p{}'.format(j, i)
            if shape[0] == 1:
                container[i, j] = Image(np.transpose(data[name][0][0])[0], spacing[1:])
            else:
                container[i, j] = Image(np.transpose(data[name][0][0]), spacing)


    return container




import os

import h5py

import miplib.ui.utils as uiutils
from miplib.data.containers.fourier_correlation_data import FourierCorrelationDataCollection
from miplib.data.containers.image import Image


class FourierCorrelationDataWriter(object):
    """
    A class for wrtiting Fourier Correlation Data into a file.
    """
    # region Constructor and Destructor
    def __init__(self, output_dir, filename, append=False):

        # Create output dir, if it doesn't exist output dir if
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        output_path = os.path.join(output_dir, filename)

        assert output_path.endswith(".hdf5")

        if os.path.isfile(output_path):
            self.data = h5py.File(output_path, mode="r+" if append else "w")
        else:
            self.data = h5py.File(output_path, mode="w")

    def __del__(self):
        self.close()
    # endregion

    def write_metadata(self, metadata):
        """
        Write a metadata dictionary to the HDF5 files header as a general description of
        the dataset.
        :param metadata: a dictionary with all the necessary data to be written in the file
        """
        assert isinstance(metadata, dict)

        for key, value in metadata:
            self.data.attrs[key] = value

    def write_images(self, images):
        """
        Write images to the data structure
        :param images: a tuple of Image objects, or a single image
        """
        if not isinstance(images, tuple):
            images = tuple(images)
        for image in images:
            assert isinstance(image, Image)

        self.data.create_group("images")

        image_name_prefix = "image_"
        for idx, image in enumerate(images):
            image_name = image_name_prefix+str(idx)
            self.data["images"].create_dataset(image_name, data=image)
            if image.ndim == 2:
                self.data["images"][image_name].attrs["pixel_size"] = "%d %d (yx)" % image.spacing
            else:
                self.data["images"][image_name].attrs["pixel_size"] = "%d %d %d (zyx)" % image.spacing

    def write_data_set(self, data):
        """
        Write Fourier Correlation Data (FRC, FSC etc) to the data structure.
        :param data:
        :type data: FourierCorrelationDataCollection
        """
        assert isinstance(data, FourierCorrelationDataCollection)

        group_prefix = "data_set_"

        for angle, data_set in data:
            group_name = group_prefix + angle
            if group_name in self.data:
                if not uiutils.get_user_input(
                        "The dataset %s already exists in the file structure. Do you want"
                        "to overwrite it?" % angle):
                    continue

                # Create a group fot every dataset and sub-groups for the two dictionaries
                # in the FourierCorrelationData structure.
                data_set_group = self.data.create_group(group_name)
                resolution_group = data_set_group.create_group("resolution")
                correlation_group = data_set_group.create_group("correlation")

                resolution_group.create_dataset("threshold", data=data_set.resolution["threshold"])
                resolution_group.attrs["resolution"] = data_set.resolution["resolution"]
                resolution_group.attrs["resolution-point"] = "%d %d (yx)" % data_set.resolution["resolution-point"]
                resolution_group.attrs["criterion"] = data_set.resolution["criterion"]
                resolution_group.create_dataset("resolution-threshold-coefficients",
                                                data=data_set.resolution["resolution-threshold-coefficients"])

                correlation_group.create_dataset("correlation", data=data_set.correlation["correlation"])
                correlation_group.create_dataset("frequency", data=data_set.correlation["frequency"])
                correlation_group.create_dataset("points-x-bin", data=data_set.correlation["points-x-bin"])
                correlation_group.create_dataset("curve-fit", data=data_set.correlation["curve-fit"])
                correlation_group.create_dataset("curve-fit-coefficients",
                                                 data=data_set.correlation["curve-fit-coefficients"])

    def close(self):
        """
        A function to explicitly close the data file (will be called by the destructor, if not.
        :return:
        """
        self.data.close()




from math import floor

import numpy as np
import SimpleITK as sitk

from miplib.data.containers.array_detector_data import ArrayDetectorData
from miplib.data.containers.image import Image
from miplib.processing import itk
from miplib.processing.registration import registration
from miplib.processing.windowing import apply_hamming_window
import miplib.processing.image as imops


def find_image_shifts(data, options, photosensor=0):
    """
    Register all images in an ISM ArrayDetectorData dataset. The central image (pixel 12)
    is used as reference and other images are aligned with it. This function used an
    iterative algorithm (ITK) that can be customized with various command line options.


    :param options: Various options that can be used on fine tune the image registration. Look into
    supertomo_options.py file
    :param data: ArrayDetectorData object with all the individual images
    :param photosensor: The photosensor number (from 0 upwards) that is to be processed
    :return: a three element tuple: x offset, y offset, transforms. The x and y offsets are expressed
    in physical units (um). The transforms are sitk.TranslationTransform objects that can be used
    to resample the images into a common coordinate system.
    """
    assert isinstance(data, ArrayDetectorData)
    assert photosensor < data.ngates

    fixed_image = itk.convert_to_itk_image(data[photosensor, int(floor(data.ndetectors / 2))])
    x = []
    y = []
    transforms = []

    for idx in range(data.ndetectors):
        image = data[photosensor, idx]
        moving_image = itk.convert_to_itk_image(image)
        transform = registration.itk_registration_rigid_2d(fixed_image, moving_image, options)
        x_new, y_new = transform.GetParameters()
        x.append(x_new)
        y.append(y_new)
        transforms.append(transform)

    return x, y, transforms


def get_theoretical_shifts(pitch=75, wavelength=550, detector_size_au=1.5):
    magnification = 500

    # Make a pure theoretical shifts grid
    pitch_pt = pitch / magnification
    axis = np.linspace(-2 * pitch_pt, 2 * pitch_pt, 5)
    y_pt, x_pt = np.meshgrid(axis, axis)

    x_pts = list(x_pt.ravel())
    y_pts = list(y_pt.ravel())[::-1]


def find_image_shifts_frequency_domain(data, photosensor=0):
    """
    Register all image in an ISM ArrayDetectorDAta dataset, with a single step frequency domain
    phase correlation based method. This might be slightly faster than the iterative method
    above (depending on the sampling strategy in the latter mainly), but usually does not
    work quite as well.

    :param data: ArrayDetectorData object with all the individual images
    :param photosensor: The photosensor number (from 0 upwards) that is to be processed
    :return: a three element tuple: x offset, y offset, transforms. The x and y offsets are expressed
    in physical units (um). The transforms are sitk.TranslationTransform objects that can be used
    to resample the images into a common coordinate system.
    """
    assert isinstance(data, ArrayDetectorData)
    assert photosensor < data.ngates

    spacing = data[0,0].spacing
    fixed_image = Image(apply_hamming_window(data[photosensor, int(floor(data.ndetectors / 2))]), spacing)
    x = []
    y = []
    transforms = []

    for idx in range(data.ndetectors):
        moving_image = Image(apply_hamming_window(data[photosensor, idx]), spacing)
        y_new, x_new = registration.phase_correlation_registration(fixed_image, moving_image,
                                                                   verbose=True, resample=False)
        tfm = sitk.TranslationTransform(2)
        tfm.SetParameters((x_new, y_new))

        x.append(x_new)
        y.append(y_new)
        transforms.append(tfm)

    return x, y, transforms


def shift_and_sum(data, transforms, photosensor=0, detectors=None, supersampling=1.0):
    """
    Adaptive ISM pixel reassignment. Please use one of the functions above to figure out
    the shifts first, if you haven't already.

    :param supersampling: Insert a number != 1, if you want to rescale the result image to
    a different size. This might make sense, if you the original sampling has been sampled
    sparsely
    :param data: ArrayDetectorData object with all the individual images
    :param transforms: ITK spatial transformation that are to be used for the resampling
    :param photosensor: The photosensor index, if more than one
    :param detectors: a list of detectors to be included in the reconstruction. If None given (default),
    all the images will be used
    :return: reconstruction result Image
    """
    assert isinstance(data, ArrayDetectorData)
    assert isinstance(transforms, list) and len(transforms) == data.ndetectors

    if supersampling != 1.0:
        new_shape = list(int(i*supersampling) for i in data[photosensor, 0].shape)
        new_spacing = list(i/supersampling for i in data[photosensor, 0].spacing)
        output = Image(np.zeros(new_shape, dtype=np.float64), new_spacing)
    else:
        output = Image(np.zeros(data[photosensor, 0].shape, dtype=np.float64), data[photosensor, 0].spacing)

    if detectors is None:
        detectors = list(range(data.ndetectors))

    for i in detectors:
        image = itk.resample_image(
            itk.convert_to_itk_image(data[photosensor, i]),
            transforms[i],
            reference=itk.convert_to_itk_image(output))

        output += itk.convert_from_itk_image(image)

    return output


def shift(data, transforms):
    """
    Resamples all the images in an ArrayDetectorData structure with the supplied transforms,
    and saves the result in a new ArrayDetectorData structure

    :param data: ArrayDetectorData object with images
    :param transforms: A list of transforms (Simple ITK), one for each image
    :return: ArrayDetectorDAta object with shifted images
    """

    assert isinstance(data, ArrayDetectorData)
    assert isinstance(transforms, list) and len(transforms) == data.ndetectors

    shifted = ArrayDetectorData(data.ndetectors, data.ngates)

    for gate in range(data.ngates):
        for i in range(data.ndetectors):
            image = itk.resample_image(
                itk.convert_to_itk_image(data[gate, i]),
                transforms[i])

            shifted[gate, i] = itk.convert_from_itk_image(image)

    return shifted


def sum(data, photosensor=0, detectors=None):
    """
    Sums all the images in a ArrayDetectorData structure

    :param detectors: A subset of detectors to be summed. If left empty, all the images
    will be summed
    :param photosensor: The photosensor index.
    :param data: ArrayDetectorData object with images
    :return: result Image
    """
    assert isinstance(data, ArrayDetectorData)

    if detectors is None:
        detectors = list(range(data.ndetectors))

    result = np.zeros(data[0,0].shape, dtype=np.float64)

    for i in detectors:
        result += data[photosensor, i]

    return Image(result, data[0, 0].spacing)




from math import pi


def degrees_to_radians(angle):
    if angle == 0:
        return 0
    else:
        return angle*pi/180


def radians_to_degrees(angle):
    if angle == 0:
        return 0
    else:
        return angle*180.0/pi


"""
A convenience script to calculate FRC for images in a directory.
I have a nicer version in a notebook -- will be updated.
"""

import datetime
import os
import sys

import numpy as np
import pandas
import miplib.analysis.resolution.fourier_ring_correlation as frc
from miplib.data.io import read as imread
import miplib.ui.cli.miplib_entry_point_options as options
import miplib.processing.to_string as strutils
import miplib.processing.image as imops

def main():

    # Get input arguments
    args = options.get_frc_script_options(sys.argv[1:])
    path = args.working_directory

    # Create output directory
    output_dir = datetime.datetime.now().strftime("%Y-%m-%d") + '_MIPLIB_output'
    output_dir = os.path.join(args.working_directory, output_dir)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    date_now = datetime.datetime.now().strftime("%H-%M-%S")

    filename = date_now + "_miplib_{}_frc_results.csv".format(args.mode)
    filename = os.path.join(output_dir, filename)

    # Get image file names, sort in alphabetic order and complete.
    files_list = list(i for i in os.listdir(path)
                      if i.endswith((".jpg", ".tif", ".tiff", ".png")))
    files_list.sort()
    print(('Number of images to analyze: ', len(files_list)))

    df_main = pandas.DataFrame()

    if args.frc_mode == "two-image":
       
        def pairwise(iterable):
            a = iter(iterable)
            return zip(a, a)

        for im1, im2 in pairwise(files_list):
            # Read images
            image1 = imread.get_image(os.path.join(path, im1))
            image2 = imread.get_image(os.path.join(path, im2))

            result = frc.calculate_two_image_frc(image1, image2, args)
            title = strutils.common_start(im1, im2)

            df_temp = pandas.DataFrame()
            df_temp['image'] = title
            df_temp['resolution'] = result.resolution['resolution']

            df_main = pandas.concat([df_main, df_temp])

    elif args.frc_mode == "one-image":
        for im in files_list:
            image = imread.get_image(os.path.join(path, im))

            result = frc.calculate_single_image_frc(image, args)

            title = im.split('.')[0]
            df_temp = pandas.DataFrame()
            df_temp['image'] = title
            df_temp['resolution'] = result.resolution['resolution']

            df_main = pandas.concat([df_main, df_temp])

    else:
        raise NotImplementedError()

    df_main.index = list(range(len(df_main)))
    df_main.to_csv(filename)

if __name__ == '__main__':
    main()



"""
Sami Koho 01/2017

Image resolution measurement by Fourier Ring Correlation.

"""

import numpy as np

import miplib.data.iterators.fourier_ring_iterators as iterators
import miplib.processing.image as imops
from miplib.data.containers.fourier_correlation_data import FourierCorrelationData, \
    FourierCorrelationDataCollection
from miplib.data.containers.image import Image
from . import analysis as fsc_analysis
from miplib.processing import windowing

def calculate_single_image_frc(image, args, average=True, trim=True, z_correction=1):
    """
    A simple utility to calculate a regular FRC with a single image input

    :param image: the image as an Image object
    :param args:  the parameters for the FRC calculation. See *miplib.ui.frc_options*
                  for details
    :return:      returns the FRC result as a FourierCorrelationData object

    """
    assert isinstance(image, Image)

    frc_data = FourierCorrelationDataCollection()

    # Hamming Windowing
    if not args.disable_hamming:
        spacing = image.spacing
        image = Image(windowing.apply_hamming_window(image), spacing)

    # Split and make sure that the images are the same siz
    image1, image2 = imops.checkerboard_split(image)
    #image1, image2 = imops.reverse_checkerboard_split(image)
    image1, image2 = imops.zero_pad_to_matching_shape(image1, image2)

    # Run FRC
    iterator = iterators.FourierRingIterator(image1.shape, args.d_bin)
    frc_task = FRC(image1, image2, iterator)
    frc_data[0] = frc_task.execute()

    if average:
        # Split and make sure that the images are the same size
        image1, image2 = imops.reverse_checkerboard_split(image)
        image1, image2 = imops.zero_pad_to_matching_shape(image1, image2)
        iterator = iterators.FourierRingIterator(image1.shape, args.d_bin)
        frc_task = FRC(image1, image2, iterator)

        frc_data[0].correlation["correlation"] *= 0.5
        frc_data[0].correlation["correlation"] += 0.5*frc_task.execute().correlation["correlation"]

    freqs = frc_data[0].correlation["frequency"].copy()
    
    def func(x, a, b, c, d):
        return a * np.exp(c * (x - b)) + d
  
    params = [0.95988146, 0.97979108, 13.90441896, 0.55146136]

    # Analyze results
    analyzer = fsc_analysis.FourierCorrelationAnalysis(frc_data, image1.spacing[0], args)

    result = analyzer.execute(z_correction=z_correction)[0]
    point = result.resolution["resolution-point"][1]

    log_correction = func(point, *params)
    result.resolution["spacing"] /= log_correction
    result.resolution["resolution"] /= log_correction

    return result

def calculate_two_image_frc(image1, image2, args, z_correction=1):
    """
    A simple utility to calculate a regular FRC with a two image input

    :param image: the image as an Image object
    :param args:  the parameters for the FRC calculation. See *miplib.ui.frc_options*
                  for details
    :return:      returns the FRC result as a FourierCorrelationData object
    """
    assert isinstance(image1, Image)
    assert isinstance(image2, Image)

    assert image1.shape == image2.shape

    frc_data = FourierCorrelationDataCollection()

    spacing = image1.spacing

    if not args.disable_hamming:

        image1 = Image(windowing.apply_hamming_window(image1), spacing)
        image2 = Image(windowing.apply_hamming_window(image2), spacing)

    # Run FRC
    iterator = iterators.FourierRingIterator(image1.shape, args.d_bin)
    frc_task = FRC(image1, image2, iterator)
    frc_data[0] = frc_task.execute()

    # Analyze results
    analyzer = fsc_analysis.FourierCorrelationAnalysis(frc_data, image1.spacing[0], args)

    return analyzer.execute(z_correction=z_correction)[0]

def calculate_single_image_sectioned_frc(image, args, rotation=45, orthogonal=True, trim=True):
    """
    A function utility to calculate a single image FRC on a Fourier ring section. The section
    is defined by the section size d_angle (in args) and the section rotation.
    :param image: the image as an Image object
    :param args:  the parameters for the FRC calculation. See *miplib.ui.frc_options*
                  for details
    :param rotation: defines the orientation of the fourier ring section
    :param orthogonal: if True, FRC is calculated from two sections, oriented at rotation
    and rotation + 90 degrees
    :return:      returns the FRC result as a FourierCorrelationData object

    """
    assert isinstance(image, Image)

    frc_data = FourierCorrelationDataCollection()

    # Hamming Windowing
    if not args.disable_hamming:
        spacing = image.spacing
        image = Image(windowing.apply_hamming_window(image), spacing)
   

    # Run FRC
    def frc_helper(image1, image2, args, rotation):
        iterator = iterators.SectionedFourierRingIterator(image1.shape, args.d_bin, args.d_angle)
        iterator.angle = rotation
        frc_task = FRC(image1, image2, iterator)
        return frc_task.execute()
    
    image1, image2 = imops.checkerboard_split(image)
    image1, image2 = imops.zero_pad_to_matching_shape(image1, image2)

    image1_r, image2_r = imops.reverse_checkerboard_split(image)
    image1_r, image2_r = imops.zero_pad_to_matching_shape(image1_r, image2_r)

    pair_1 = frc_helper(image1, image2, args, rotation)
    pair_2 = frc_helper(image1_r, image2_r, args, rotation)
    
    pair_1.correlation["correlation"] * 0.5
    pair_1.correlation["correlation"] += 0.5 * pair_2.correlation["correlation"]
   
    if orthogonal:
        pair_1_o = frc_helper(image1, image2, args, rotation+90)
        pair_2_o = frc_helper(image1_r, image2_r, args, rotation+90)
    
        pair_1_o.correlation["correlation"] * 0.5
        pair_1_o.correlation["correlation"] += 0.5 * pair_2_o.correlation["correlation"]

        pair_1.correlation["correlation"] += 0.5 * pair_1_o.correlation["correlation"]
    
    frc_data[0] = pair_1

    freqs = frc_data[0].correlation["frequency"].copy()
    
    def func(x, a, b, c, d):
        return a * np.exp(c * (x - b)) + d
  
    params = [0.95988146, 0.97979108, 13.90441896, 0.55146136]

    # Analyze results
    analyzer = fsc_analysis.FourierCorrelationAnalysis(frc_data, image1.spacing[0], args)

    result = analyzer.execute()[0]
    point = result.resolution["resolution-point"][1]

    log_correction = func(point, *params)
    result.resolution["spacing"] /= log_correction
    result.resolution["resolution"] /= log_correction

    return result

class FRC(object):
    """
    A class for calcuating 2D Fourier ring correlation. Contains
    methods to calculate the FRC as well as to plot the results.
    """

    def __init__(self, image1, image2, iterator):
        assert isinstance(image1, Image)
        assert isinstance(image2, Image)

        if image1.shape != image2.shape or tuple(image1.spacing) != tuple(image2.spacing):
            raise ValueError("The image dimensions do not match")
        if image1.ndim != 2:
            raise ValueError("Fourier ring correlation requires 2D images.")

        self.pixel_size = image1.spacing[0]

        # Expand to square
        image1 = imops.zero_pad_to_cube(image1)
        image2 = imops.zero_pad_to_cube(image2)

        self.iterator = iterator
        # Calculate power spectra for the input images.
        self.fft_image1 = np.fft.fftshift(np.fft.fft2(image1))
        self.fft_image2 = np.fft.fftshift(np.fft.fft2(image2))

        # Get the Nyquist frequency
        self.freq_nyq = int(np.floor(image1.shape[0] / 2.0))

    def execute(self):
        """
        Calculate the FRC
        :return: Returns the FRC results.

        """
        radii = self.iterator.radii
        c1 = np.zeros(radii.shape, dtype=np.float32)
        c2 = np.zeros(radii.shape, dtype=np.float32)
        c3 = np.zeros(radii.shape, dtype=np.float32)
        points = np.zeros(radii.shape, dtype=np.float32)

        for ind_ring, idx in self.iterator:
            subset1 = self.fft_image1[ind_ring]
            subset2 = self.fft_image2[ind_ring]
            c1[idx] = np.sum(subset1 * np.conjugate(subset2)).real
            c2[idx] = np.sum(np.abs(subset1) ** 2)
            c3[idx] = np.sum(np.abs(subset2) ** 2)

            points[idx] = len(subset1)

        # Calculate FRC
        spatial_freq = radii.astype(np.float32) / self.freq_nyq
        n_points = np.array(points)

        with np.errstate(divide="ignore", invalid="ignore"):
            frc = np.abs(c1) / np.sqrt(c2 * c3)
            frc[frc == np.inf] = 0.0
            frc = np.nan_to_num(frc)


        data_set = FourierCorrelationData()
        data_set.correlation["correlation"] = frc
        data_set.correlation["frequency"] = spatial_freq
        data_set.correlation["points-x-bin"] = n_points

        return data_set


import os
import pandas as pd

from . import filters
from miplib.data.io import read
import miplib.analysis.resolution.fourier_ring_correlation as frc

def evaluate_image_quality(image, options):
  """
  Calculate quality features sof a single image
  """
  
  # Run spatial domain analysis
  task = filters.LocalImageQuality(image, options)
  task.set_smoothing_kernel_size(100)
  entropy = task.calculate_image_quality()

  # Run frequency domain analysis
  task2 = filters.FrequencyQuality(image, options)
  results = task2.analyze_power_spectrum()

  task3 = filters.SpectralMoments(image, options)
  moments = task3.calculate_spectral_moments()

  task4 = filters.BrennerImageQuality(image, options)
  brenner = task4.calculate_brenner_quality()

  # Save results
  results.insert(0, moments)
  results.insert(0, brenner)
  results.insert(0, entropy)
    
  return results


def batch_evaluate_image_quality(path, options):
    """
    Batch calculate quality features for images in a directory
    :param options: options for the quality ranking scripts, as in miplib/ui/image_quality_options.py
    :parame path:   directory that contains the images to be analyzed
    """

    df = pd.DataFrame(columns=["Filename", "tEntropy", "tBrenner", "fMoments", "fMean", "fSTD", "fEntropy",
               "fTh", "fMaxPw", "Skew", "Kurtosis", "MeanBin", "Resolution"])

    for idx, image_name in enumerate(os.listdir(path)):
        if options.file_filter is None or options.file_filter in image_name:
            real_path = os.path.join(path, image_name)
            # Only process images
            if not os.path.isfile(real_path) or not real_path.endswith((".jpg", ".tif", ".tiff", ".tif")):
                continue
            # ImageJ files have particular TIFF tags that can be processed correctly
            # with the options.imagej switch
            image = read.get_image(real_path, channel=options.rgb_channel)

            # Only grayscale images are processed. If the input is an RGB image,
            # a channel can be chosen for processing.
            results = evaluate_image_quality(image, options)
            results.insert(0, real_path)

            # Add resolution value to the end
            results.append(frc.calculate_single_image_frc(image, options).resolution["resolution"])

            df.loc[idx] = results

            print ("Done analyzing {}".format(image_name))

    return df
        


import argparse

def get_deconvolution_options_group(parser):
    assert isinstance(parser, argparse.ArgumentParser)
    group = parser.add_argument_group("Deconvolution", "Options for controlling the deconvolution algorithm")

    group.add_argument(
        '--max-nof-iterations',
        type=int,
        default=20,
        help='Specify maximum number of iterations.'
    )
    group.add_argument(
        '--update-blind-psf',
        type=int,
        default=0,
        help='Updates the blind psf during deconvolution, at specified iteration intervals.'
    )
    group.add_argument(
        '--convergence-epsilon',
        dest='convergence_epsilon',
        type=float,
        default=0.05,
        help='Specify small positive number that determines '
             'the window for convergence criteria.'
    )

    group.add_argument(
        '--wiener-nsr',
        type=float,
        default=100.0
    )

    group.add_argument(
        '--first-estimate',
        choices=['image',
                 'blurred',
                 'image_mean',
                 'constant'],
        default='image',
        help='Specify first estimate for iteration.'
    )

    group.add_argument(
        '--estimate-constant',
        dest='estimate_constant',
        type=float,
        default=1.0
    )

    group.add_argument(
        '--save-intermediate-results',
        action='store_true',
        help='Save intermediate results.'
    )

    group.add_argument(
        '--output-cast',
        dest='output_cast',
        action='store_true',
        help='By default the fusion output is returned as a 32-bit image'
             'This switch can be used to enable 8-bit unsigned output'
    )

    group.add_argument(
        '--blocks',
        dest='num_blocks',
        type=int,
        default=1,
        help="Define the number of blocks you want to break the images into"
             "for the image fusion. This argument defaults to 1, which means"
             "that the entire image will be used -- you should define a larger"
             "number to optimize memory consumption"
    )
    group.add_argument(
        '--stop-tau',
        type=float,
        default=0.0001,
        help='Specify parameter for tau-stopping criteria.'
    )
    group.add_argument(
        '--tv-lambda',
        type=float,
        default=0,
        help="Enable Total Variation regularization by selecting value > 0"
    )

    group.add_argument(
        '--pad',
        dest='block_pad',
        type=int,
        default=0,
        help='The amount of padding to apply to a fusion block.'
    )

    group.add_argument(
        '--memmap-estimates',
        action='store_true'
    )

    group.add_argument(
        '--disable-tau1',
        action='store_true'
    )

    group.add_argument(
        '--disable-fft-psf-memmap',
        action='store_true'
    )

    group.add_argument(
        '--rl-background',
        type=float,
        default=0.0,
        help="Background correction term for RL"
    )
    group.add_argument(
        '--rl-auto-background',
        action="store_true"
    )

    return parser








# coding=utf-8
"""
File:        filters.py
Author:      Sami Koho (sami.koho@gmail.com)

Description:
This file contains the filters that are used for calculating the
image quality parameters in the PyImageQuality software.
-   The LocalImageQuality class is used to run spatial domain
    analysis. It calculates the Shannon entropy value at a
    masked part of an image
-   The FrequencyQuality class is used to calculate statistical
    quality parameters in the frequency domain. The calculations
    are based on the analysis of the tail of the 1D power spect-
    rum.
-   Brenner and Spectral domain autofocus metrics were impelemnted
    as well, based on the two classes above.
"""

import argparse
from math import floor

import numpy as np
from matplotlib import pyplot as plt
from scipy import ndimage, fftpack, stats

from . import utils
from miplib.data.containers.image import Image
from miplib.processing import image as imutils
from miplib.data.iterators.fourier_ring_iterators import FourierRingIterator


def get_common_options(parser):
    """
    Common command-line options for the image-quality filters
    """
    assert isinstance(parser, argparse.ArgumentParser)
    group = parser.add_argument_group(
        "Filters common", "Common options for the quality filters"
    )
    group.add_argument(
        "--power-averaging",
        dest="power_averaging",
        choices=["radial", "additive"],
        default="additive"
    )
    group.add_argument(
        "--normalize-power",
        dest="normalize_power",
        action="store_true"

    )
    group.add_argument(
        "--use-mask",
        dest="use_mask",
        action="store_true"
    )
    group.add_argument(
        "--invert-mask",
        dest="invert_mask",
        action="store_true"
    )
    group.add_argument(
        "--power-threshold",
        dest="power_threshold",
        type=float,
        default=0.4
    )
    group.add_argument(
        "--spatial-threshold",
        dest="spatial_threshold",
        type=int,
        default=80
    )
    group.add_argument(
        "--show-plots",
        dest="show_plots",
        action="store_true"
    )
    return parser


class Filter(object):
    """
    A base class for a filter utilizing Image class object
    """
    def __init__(self, image, options, physical=False, verbal=False):

        assert isinstance(image, Image)
        self.options = options

        self.data = image

        self.spacing = self.data.spacing
        self.dimensions = self.data.shape
        self.physical = physical
        self.verbal = verbal

    def set_physical_coordinates(self):
        self.physical = True

    def set_pixel_coordinates(self):
        self.physical = False


class LocalImageQuality(Filter):
    """
    This is a filter for quantifying  image quality, based on the calculation
    of Shannon entropy at image neighborhoods that contain the highest amount
    of detail.
    """

    def __init__(self, image, options, physical=False, verbal=False):

        Filter.__init__(self, image, options, physical, verbal)

        self.data_temp = None
        self.kernel_size = []

    def set_smoothing_kernel_size(self, size):

        if isinstance(size, list):
            assert len(size) == len(self.spacing)
            sizes = size
        elif isinstance(size, float) or isinstance(size, int):
            sizes = [size, ] * len(self.spacing)
        else:
            print("Unknown size type")
            return

        if self.physical is True:
            for i in range(len(sizes)):
                self.kernel_size[i] = sizes[i]/self.spacing[i]
                assert self.kernel_size[i] < self.dimensions[i]
        else:
            self.kernel_size = sizes
            assert all(x < y for x, y in zip(self.kernel_size, self.dimensions)), \
                "Kernel can not be larger than image"

    def run_mean_smoothing(self, return_result=False):
        """
        Mean smoothing is used to create a mask for the entropy calculation
        """

        assert len(self.kernel_size) == len(self.dimensions)
        self.data_temp = ndimage.uniform_filter(self.data[:], size=self.kernel_size)

        if return_result:
            return Image(self.data_temp, self.spacing)

    def calculate_entropy(self):
        """
        Returns the Shannon entropy value of an image.
        """
        # Calculate histogram
        histogram = ndimage.histogram(
            self.data_temp,
            self.data_temp.min(),
            self.data_temp.max(), 50
        )
        # Exclude zeros
        histogram = histogram[np.nonzero(histogram)]
        # Normalize histogram bins to sum to one
        histogram = histogram.astype(float)/histogram.sum()
        return -np.sum(histogram*np.log2(histogram))

    def find_sampling_positions(self):
        """
        Create a mask by finding pixel positions in the smoothed image
        that have pixel values higher than 80% of the maximum value.
        """
        peaks = np.percentile(self.data_temp, self.options.spatial_threshold)
        mask = np.where(self.data_temp >= peaks, 1, 0)
        if self.options.invert_mask:
            return np.invert(mask.astype(bool))
        else:
            return mask

    def calculate_image_quality(self, kernel=None, show=False):
        """
        Calculate an estimate for image quality, based on the
        Shannon entropy measure. options.use_mask switch can
        be used to limit the entropy calculation to detailed
        parts of the image.
        """
        if self.options.use_mask:
            if kernel is not None:
                self.set_smoothing_kernel_size(kernel)

            assert len(self.kernel_size) != 0

            if self.data_temp is None:
                self.run_mean_smoothing()

            positions = self.find_sampling_positions()
            self.data_temp = self.data[:][np.nonzero(positions)]
            if show:
                Image(self.data[:]*positions, self.spacing).show()
        else:
            self.data_temp = self.data[:]

        return self.calculate_entropy()


class FrequencyQuality(Filter):
    """
    A filter for calculated image-quality related parameters in the frequency
    domain. First a one-dimensional power spectrum is calculated for an image,
    after which various types of statistics are calculated for the power
    spectrum tail (frequencies > 40% of Nyquist)
    """
    def __init__(self, image, options, physical=False, verbal=False):

        Filter.__init__(self, image, options, physical=physical, verbal=verbal)

        # Additive form of power spectrum calculation requires a square shaped
        # image
        if self.options.power_averaging == "additive":
            self.data = imutils.crop_to_largest_square(self.data)

        self.simple_power = None
        self.power = None
        self.kernel_size = []

    def set_image(self, image):
        self.data = image

    def calculate_power_spectrum(self):
        """
        A function that is used to calculate a centered 2D power spectrum.
        Additionally the power spectrum can be normalized by image dimensions
        and image intensity mean, if necessary.
        """
        self.power = np.abs(np.fft.fftshift(np.fft.fft2(self.data[:])))**2
        if self.options.normalize_power:
            dims = self.data[:].shape[0]*self.data[:].shape[1]
            mean = np.mean(self.data[:])
            self.power /= (dims*mean)

    def calculate_radial_average(self, bin_size=2):
        """
        Convert a 2D centered power spectrum into 1D by averaging spectral
        power at different radiuses from the zero frequency center
        """
        iterator = FourierRingIterator(self.power.shape, d_bin=bin_size)

        average = np.zeros(iterator.nbins)

        for idx, ring in enumerate(iterator):
            subset = self.power[ring]
            average[idx] = float(subset.sum())/subset.size

        dx = self.data.spacing[0]
        f_k = np.linspace(0, 1, iterator.nbins) * (1.0/(2*dx))

        self.simple_power = [f_k, average]

        if self.options.show_plots:
            plt.plot(np.log10(self.simple_power[0]))
            plt.ylabel("Average power")
            plt.xlabel("Frequency")
            plt.show()

    def calculate_summed_power(self):
        """
        Calculate a 1D power spectrum fro 2D power spectrum, by summing all rows and
        columns, and then summing negative and positive frequencies, to form a
        N/2+1 long 1D array. This approach is significantly faster to calculate
        than the radial average.
        """

        power_sum = np.zeros(self.power.shape[0])
        for i in range(len(self.power.shape)):
            power_sum += np.sum(self.power, axis=i)
        zero = floor(float(power_sum.size)/2)
        power_sum[zero+1:] = power_sum[zero+1:]+power_sum[:zero-1][::-1]
        power_sum = power_sum[zero:]
        dx = self.data.spacing[0]
        f_k = np.linspace(0, 1, power_sum.size)*(1.0/(2*dx))

        self.simple_power = [f_k, power_sum]

        if self.options.show_plots:
            plt.plot(self.simple_power[0], self.simple_power[1], linewidth=2, color="red")
            plt.ylabel("Total power")
            plt.yscale('log')
            plt.xlabel('Frequency')
            plt.show()

    def analyze_power_spectrum(self):
        """
        Run the image quality analysis on the power spectrum
        """
        assert self.data is not None, "Please set an image to process"
        self.calculate_power_spectrum()

        # Choose a method to calculate 1D power spectrum
        if self.options.power_averaging == "radial":
            self.calculate_radial_average()
        elif self.options.power_averaging == "additive":
            self.calculate_summed_power()
        else:
            raise NotImplementedError

        # Extract the power spectrum tail
        hf_sum = self.simple_power[1][self.simple_power[0] > self.options.power_threshold*self.simple_power[0].max()]

        # Calculate parameters
        f_th = self.simple_power[0][self.simple_power[0] > self.options.power_threshold*self.simple_power[0].max()][-utils.analyze_accumulation(hf_sum, .2)]
        mean = np.mean(hf_sum)
        std = np.std(hf_sum)
        entropy = utils.calculate_entropy(hf_sum)
        nm_th = 1.0e9/f_th
        pw_at_high_f = np.mean(self.simple_power[1][self.simple_power[0] > .9*self.simple_power[0].max()])
        skew = stats.skew(np.log(hf_sum))
        kurtosis = stats.kurtosis(hf_sum)
        mean_bin = np.mean(hf_sum[0:5])

        return [mean, std, entropy, nm_th, pw_at_high_f, skew, kurtosis, mean_bin]

    def show_all(self):
        """
        A small utility to show a plot of the 2D and 1D power spectra
        """
        fig, subplots = plt.subplots(1, 2)
        if self.power is not None:
            subplots[0].imshow(np.log10(self.power))
        if self.simple_power is not None:
            #index = int(len(self.simple_power[0])*.4)
            #subplots[1].plot(self.simple_power[0][index:], self.simple_power[1][index:], linewidth=1)
            subplots[1].plot(self.simple_power[0], self.simple_power[1], linewidth=1)
            subplots[1].set_yscale('log')
        plt.show()

    def get_power_spectrum(self):
        """
        Returns the calculated 1D power spectrum. Please make sure to create
        power spectrum before calling this.
        """
        assert self.simple_power is not None
        return self.simple_power


class SpectralMoments(FrequencyQuality):
    """
    Our implementation of the Spectral Moments autofocus metric
    Firestone, L. et al (1991). Comparison of autofocus methods for automated
    microscopy. Cytometry, 12(3), 195–206.
    http://doi.org/10.1002/cyto.990120302
    """

    def calculate_percent_spectrum(self):
        self.simple_power[1] /= (self.simple_power[1].sum()/100)

    def calculate_spectral_moments(self):
        """
        Run the image quality analysis on the power spectrum
        """
        assert self.data is not None, "Please set an image to process"
        self.calculate_power_spectrum()

        # Choose a method to calculate 1D power spectrum
        if self.options.power_averaging == "radial":
            self.calculate_radial_average()
        elif self.options.power_averaging == "additive":
            self.calculate_summed_power()
        else:
            raise NotImplementedError

        self.calculate_percent_spectrum()

        bin_index = np.arange(1, self.simple_power[1].shape[0]+1)

        return (self.simple_power[1]*np.log10(bin_index)).sum()


class BrennerImageQuality(Filter):
    """
    Our implementation of the Brenner autofocus metric
    Brenner, J. F. et al (1976). An automated microscope for cytologic research
    a preliminary evaluation. Journal of Histochemistry & Cytochemistry, 24(1),
    100–111. http://doi.org/10.1177/24.1.1254907
    """
    def __init__(self, image, options, physical=False, verbal=False):

        Filter.__init__(self, image, options, physical, verbal)
        # This is not really necessary. It was just added in order to compare
        # with the frequency domain measures.
        self.data = imutils.crop_to_largest_square(image)

    def calculate_brenner_quality(self):
        data = self.data
        rows = data.shape[0]
        columns = data.shape[1]-2
        temp = np.zeros((rows, columns))

        temp[:] = ((data[:, 0:-2] - data[:, 2:])**2)

        return temp.sum()



















import sys
import os
import numpy as np

from matplotlib import pyplot as plt
import miplib.ui.plots.scatter as scatter
import miplib.ui.plots.frc as frcplots

from miplib.data.io import carma
from miplib.data.containers.image import Image
import miplib.processing.image as imops
from miplib.processing.ism import reconstruction as ismrec
from miplib.processing.deconvolution import deconvolve, wiener
from miplib.psf import frc_psf
from miplib.data.messages import image_writer_wrappers as imwrap
from miplib.processing.segmentation import masking
import miplib.ui.cli.miplib_entry_point_options as options

from miplib.data.io.write import image as imwrite


def main():
    args = options.get_ism_script_options(sys.argv[1:])
    
    # Only allow translation transform, as rotation is not needed here, 
    args.reg_translate_only = True

    # Get Carma mat files in the working direcory
    root_dir = args.directory
    files = tuple(f for f in os.listdir(root_dir) if f.endswith(".mat"))

    # Generate ISM result for each file
    for filename in files:
        filename_prefix = filename.split(".mat")[0]

        full_path = os.path.join(root_dir, filename)

        data = carma.read_carma_mat(full_path)
        
        print ("Opened a Carma file {} for ISM reconstruction".format(filename))
        print ("The file contains images from {} laser gates and {} detector " 
               "channels.".format(data.ngates, data.ndetectors))
        print ("The image size is {} and the pixel spacing {} um".format(data[0,0].shape, data[0,0].spacing))

        print ("Calculating image shifts for adaptive reassignment.")
        # Run registration
        x,y,transforms = ismrec.find_image_shifts(data, args)

        # Save shift scatter, if saving plots is enabled
        if args.save_plots:
            fig = scatter.xy_scatter_plot_with_labels(
                y,x, range(len(x)), size=args.plot_size)
            scatter_name = "{}_shifts_scatter.eps".format(filename_prefix)
            scatter_path = os.path.join(root_dir, scatter_name)
            fig.savefig(scatter_path, dpi=1200, bbox_inches='tight', 
                pad_inches=0, transparent=True)

        # Generate ISM and confocal results
        ism_result = ismrec.shift_and_sum(data, transforms)
        result_sum = ismrec.sum(data)

        ism_result_name = "{}_apr_ism.tif".format(filename_prefix)
        confocal_result_name = "{}_open.tif".format(filename_prefix)

        imwrite(os.path.join(root_dir, ism_result_name), 
            imops.rescale_to_8_bit(ism_result))
        imwrite(os.path.join(root_dir, confocal_result_name), 
            imops.rescale_to_8_bit(result_sum))

        # Run FRC based Wiener filter, if requested
        if any(args.ism_mode == mode for mode in ("wiener", "all")):
            print ("Running the blind Wiener filter.") 
            psf = frc_psf.generate_frc_based_psf(ism_result, args)
            wiener_result = wiener.wiener_deconvolution(
                ism_result, psf, snr=args.wiener_nsr
                )
            wiener_result_name = "{}_apr_ism_bplus_wiener.tif".format(filename_prefix)
            imwrite(os.path.join(root_dir, wiener_result_name), 
            imops.rescale_to_8_bit(wiener_result))

        # Run FRC based RL deconvolution, if requested
        if any(args.ism_mode == mode for mode in ("rl", "all")): 
            print("Running the blind RL deconvolution.")
            psf = frc_psf.generate_frc_based_psf(ism_result, args)

            # Enable automatic background correction with --rl-auto-background
            if args.rl_auto_background:
                background_mask= masking.make_local_intensity_based_mask(
                    ism_result, 40, invert=True)
                masked_image = Image(ism_result*background_mask, ism_result.spacing)
                args.rl_background = np.mean(masked_image[masked_image > 0])

            # If saving intermediate results, a temp directory is created here
            if args.save_intermediate_results:
                temp_dir = os.path.join(root_dir, "{}_temp".format(filename_prefix))
                if not os.path.exists(temp_dir):
                    os.mkdir(temp_dir)
                writer = imwrap.TiffImageWriter(temp_dir)
            else:
                writer = None

            task = deconvolve.DeconvolutionRL(ism_result, psf, writer, args)
            task.execute()
            rl_result = task.get_8bit_result()

            rl_results_name = "{}_apr_ism_bplus_rl_{}.tif".format(
                filename_prefix, task.iteration_count)
            imwrite(os.path.join(root_dir, rl_results_name), 
            imops.rescale_to_8_bit(rl_result))
        
        print ("Done.")
        print()



if __name__ == "__main__":
    main()



import numpy as np
import scipy.ndimage as ndimage
import scipy.optimize as optimize
from scipy.interpolate import interp1d, UnivariateSpline
from scipy.signal import medfilt, savgol_filter
import miplib.processing.ndarray as arrayutils
from miplib.data.containers.fourier_correlation_data import FourierCorrelationDataCollection, FourierCorrelationData
import miplib.processing.converters as converters


def fit_frc_curve(data_set, degree, fit_type='spline'):
    """
    Calculate a least squares curve fit to the FRC Data
    :return: None. Will modify the frc argument in place
    """
    assert isinstance(data_set, FourierCorrelationData)

    data = data_set.correlation["correlation"]

    if fit_type == 'smooth-spline':
        equation = UnivariateSpline(data_set.correlation["frequency"],
                                    data)
        equation.set_smoothing_factor(0.25)
        # equation = interp1d(data_set.correlation["frequency"],
        #                     data, kind='slinear')

    elif fit_type == 'spline':
        equation = interp1d(data_set.correlation["frequency"],
                            data, kind='slinear')

    elif fit_type == 'polynomial':

        coeff = np.polyfit(data_set.correlation["frequency"],
                           data,
                           degree,
                           w=1 - data_set.correlation["frequency"] ** 3)
        equation = np.poly1d(coeff)
    else:
        raise AttributeError(fit_type)

    data_set.correlation["curve-fit"] = equation(data_set.correlation["frequency"])

    return equation


def calculate_snr_threshold_value(points_x_bin, snr):
    """
    A function to calculate a SNR based resolution threshold, as described
    in ...

    :param points_x_bin: a 1D Array containing the numbers of points at each
    FRC/FSC ring/shell
    :param snr: the expected SNR value
    :return:
    """
    nominator = snr + arrayutils.safe_divide(
            2.0 * np.sqrt(snr) + 1,
            np.sqrt(points_x_bin)
        )
    denominator = snr + 1 + arrayutils.safe_divide(
        2.0 * np.sqrt(snr),
        np.sqrt(points_x_bin)
    )
    return arrayutils.safe_divide(nominator, denominator)



def calculate_resolution_threshold_curve(data_set, criterion, threshold, snr):
    """
    Calculate the two sigma curve. The FRC should be run first, as the results of the two sigma
    depend on the number of points on the fourier rings.

    :return:  Adds the
    """
    assert isinstance(data_set, FourierCorrelationData)

    points_x_bin = data_set.correlation["points-x-bin"]

    if points_x_bin[-1] == 0:
        points_x_bin[-1] = points_x_bin[-2]

    if criterion == 'one-bit':
        nominator = 0.5 + arrayutils.safe_divide(
            2.4142,
            np.sqrt(points_x_bin)
        )
        denominator = 1.5 + arrayutils.safe_divide(
            1.4142,
            np.sqrt(points_x_bin)
        )
        points = arrayutils.safe_divide(nominator, denominator)

    elif criterion == 'half-bit':
        nominator = 0.2071 + arrayutils.safe_divide(
            1.9102,
            np.sqrt(points_x_bin)
        )
        denominator = 1.2071 + arrayutils.safe_divide(
            0.9102,
            np.sqrt(points_x_bin)
        )
        points = arrayutils.safe_divide(nominator, denominator)

    elif criterion == 'three-sigma':
        points = arrayutils.safe_divide(np.full(points_x_bin.shape, 3.0), (np.sqrt(points_x_bin) + 3.0 - 1))


    elif criterion == 'fixed':
        points = threshold * np.ones(len(data_set.correlation["points-x-bin"]))
    elif criterion == 'snr':
        points = calculate_snr_threshold_value(points_x_bin, snr)

    else:
        raise AttributeError()

    if criterion != 'fixed':
        #coeff = np.polyfit(data_set.correlation["frequency"], points, 3)
        #equation = np.poly1d(coeff)
        equation = interp1d(data_set.correlation["frequency"], points, kind='slinear')
        curve = equation(data_set.correlation["frequency"])
    else:
        curve = points
        equation = None

    data_set.resolution["threshold"] = curve
    return equation


class FourierCorrelationAnalysis(object):
    def __init__(self, data, spacing, args):

        assert isinstance(data, FourierCorrelationDataCollection)

        self.data_collection = data
        self.args = args
        self.spacing = spacing

    def execute(self, z_correction=1):
        """
        Calculate the spatial resolution as a cross-section of the FRC and Two-sigma curves.

        :return: Returns the calculation results. They are also saved inside the class.
                 The return value is just for convenience.
        """

        criterion = self.args.resolution_threshold_criterion
        threshold = self.args.resolution_threshold_value
        snr = self.args.resolution_snr_value
        tolerance = self.args.resolution_point_sigma
        degree = self.args.frc_curve_fit_degree
        fit_type = self.args.frc_curve_fit_type
        verbose = self.args.verbose

        def pdiff1(x):
            return abs(frc_eq(x) - two_sigma_eq(x))

        def pdiff2(x):
            return abs(frc_eq(x) - threshold)

        def first_guess(x, y, threshold):
            #y_smooth = savgol_filter(y, 5, 2)
            #return x[np.argmin(np.abs(y_smooth - threshold))]

            difference = y - threshold

            return x[np.where(difference <= 0)[0][0] - 1]
            #return x[np.argmin(np.abs(y - threshold))]

        for key, data_set in self.data_collection:

            if verbose:
                print("Calculating resolution point for dataset {}".format(key))
            frc_eq = fit_frc_curve(data_set, degree, fit_type)
            two_sigma_eq = calculate_resolution_threshold_curve(data_set, criterion, threshold, snr)

            """
            Todo: Make the first quess adaptive. For example find the data point at which FRC
            value is closest to the mean of the threshold
            """


            # Find intersection
            fit_start = first_guess(data_set.correlation["frequency"],
                                    data_set.correlation["correlation"],
                                    np.mean(data_set.resolution["threshold"])
            )
            if self.args.verbose:
                print("Fit starts at {}".format(fit_start))
                disp = 1
            else:
                disp = 0
            root = optimize.fmin(pdiff2 if criterion == 'fixed' else pdiff1, fit_start, disp=disp)[0]
            data_set.resolution["resolution-point"] = (frc_eq(root), root)
            data_set.resolution["criterion"] = criterion

            angle = converters.degrees_to_radians(int(key))
            z_correction_multiplier = (1+(z_correction-1)*np.abs(np.sin(angle)))
            resolution =z_correction_multiplier*(2*self.spacing / root)

            data_set.resolution["resolution"] = resolution
            data_set.resolution["spacing"] = self.spacing*z_correction_multiplier

            self.data_collection[int(key)] = data_set


            # # # Find intersection
            # root, result = optimize.brentq(
            #     pdiff2 if criterion == 'fixed' else pdiff1,
            #     0.0, 1.0, xtol=tolerance, full_output=True)
            #
            # # Save result, if intersection was found
            # if result.converged is True:
            #     data_set.resolution["resolution-point"] = (frc_eq(root), root)
            #     data_set.resolution["criterion"] = criterion
            #     resolution = 2 * self.spacing / root
            #     data_set.resolution["resolution"] = resolution
            #     self.data_collection[int(key)] = data_set
            # else:
            #     print "Could not find an intersection for the curves for the dataset %s." % key

        return self.data_collection



import os

import h5py
import numpy
import scipy.ndimage as ndimage

import miplib.processing.itk as itkutils
import miplib.ui.utils as uiutils
from miplib.data.containers.image import Image
from miplib.data.definitions import *
import miplib.processing.ndarray as arrayutils


class ImageData(object):
    """
    The data storage in miplib is based on a HDF5 file format. This
    allows the efficient handing of large datasets
    """

    def __init__(self, path):

        if not os.path.exists(os.path.dirname(path)):
            os.mkdir(os.path.dirname(path))
        assert path.endswith(".hdf5")

        if os.path.exists(path):
            self.data = h5py.File(path, mode="r+")
            self.series_count = self.data.attrs["series_count"]
            self.channel_count = self.data.attrs["channel_count"]
        else:
            self.data = h5py.File(path, mode="w")
            self.series_count = 0
            self.channel_count = 1

            self.data.attrs["series_count"] = self.series_count
            self.data.attrs["channel_count"] = self.channel_count

        self.active_image = None

    def add_original_image(self, data, scale, index, channel, angle, spacing, chunk_size=None):
        """
        Add a source image to the HDF5 file.

        Parameters
        ----------
        :param data:            Contains an image from a single rotation angle.
                                The data should be in numpy.ndarray format.
                                Multi-channel data is expected to be a 4D array
                                in which the color channel is the first
                                dimension.
        :param scale            Percentage from full size. It is possible to save
                                multiple versions of an image in different sizes.
        :param index            The image ordering index
        :param channel          The color channel to be associated with the image.
        :param angle:           Estimated rotation angle of the view, in
                                respect to the regular STED angle
        :param spacing:         Voxel size

        :param chunk_size:      A specific chunk size can be defined here in
                                order to optimize the data access, when
                                working with partial images.
        :return:
        """
        assert isinstance(data, numpy.ndarray), "Invalid data format."

        # if int(channel) > self.channel_count + 1:
        #     raise ValueError("Add the color channels in the correct order")

        # Create a new image group, based on the ordering index. If the
        # group exists, an attempt is made to add a new dataset.
        group_name = "original/" + str(index)
        if group_name not in self.data:
            image_group = self.data.create_group(group_name)
            self.series_count += 1
            self.data.attrs["series_count"] = self.series_count
        else:
            image_group = self.data[group_name]

        name = "channel_" + str(channel) + "_scale_" + str(scale)

        # Don't overwrite an existing image.
        if name in image_group:
            return

        # Zoom axial dimension for isotropic pixel size.
        if data.ndim == 3 and spacing[0] != spacing[1]:
            print("Image index %s needs to be resampled for isotropic spacing." \
                  "This will take a minute" % index)
            z_zoom = spacing[0] / spacing[1]
            data = ndimage.zoom(data, (z_zoom, 1, 1), order=3)
            spacing = tuple(spacing[x] if x != 0 else spacing[x]/z_zoom for x in range(len(spacing)))

        # Activate chunked storage of requested
        if chunk_size is None:
            image_group.create_dataset(name, data=data)
        else:
            image_group.create_dataset(name, data=data, chunks=chunk_size)

        image_group[name].attrs["angle"] = angle
        image_group[name].attrs["spacing"] = spacing
        image_group[name].attrs["size"] = data.shape

        # # The first image is the same in the registered group as well,
        # # so a soft link will be created here.
        # if int(index) == 0:
        #     reg_group_name = "registered/" + index
        #     reg_group = self.data.create_group(reg_group_name)
        #     reg_group[name] = image_group[name]

    def add_registered_image(self, data, scale, index, channel, angle, spacing, chunk_size=None):
        """
        Add a registered/resampled image to the HDF5 file.

        Parameters
        ----------
        :param data:            Contains an image from a single rotation angle.
                                The data should be in numpy.ndarray format.
                                Multi-channel data is expected to be a 4D array
                                in which the color channel is the first
                                dimension.
        :param scale            Percentage from full size. It is possible to save
                                multiple versions of an image in different sizes.
        :param index            The image ordering index
        :param channel          The color channel to be associated with the image.
        :param angle:           Estimated rotation angle of the view, in
                                respect to the regular STED angle
        :param spacing:         Voxel size

        :param chunk_size:      A specific chunk size can be defined here in
                                order to optimize the data access, when
                                working with partial images.
        :return:
        """
        assert isinstance(data, numpy.ndarray), "Invalid data format."

        group_name = "registered/" + str(index)
        if group_name not in self.data:
            image_group = self.data.create_group(group_name)
        else:
            image_group = self.data[group_name]

        # if channel > 0:
        #     if self.channel_count == 1:
        #         raise ValueError("Invalid channel count")

        name = "channel_" + str(channel) + "_scale_" + str(scale)
        if name in image_group:
            if uiutils.get_user_input("The dataset %s already exists in image "
                                      "group %s. Do you want to overwrite "
                                      "it? " % (name, group_name)):
                del image_group[name]
            else:
                return

        if chunk_size is None:
            image_group.create_dataset(name, data=data)
        else:
            image_group.create_dataset(name, data=data, chunks=chunk_size)

        # Each image has its own attributes
        image_group[name].attrs["angle"] = angle
        image_group[name].attrs["spacing"] = spacing
        image_group[name].attrs["size"] = data.shape

    def add_psf(self, data, scale, index, channel, angle, spacing, chunk_size=None,
                calculated=False):
        """
        Add a PSF image to the HDF5 file.

        Parameters
        ----------
        :param data:            Contains an image from a single rotation angle.
                                The data should be in numpy.ndarray format.
                                Multi-channel data is expected to be a 4D array
                                in which the color channel is the first
                                dimension.
        :param scale            Percentage from full size. It is possible to save
                                multiple versions of an image in different sizes.
        :param index            The image ordering index
        :param channel          The color channel to be associated with the image.
        :param angle:           Estimated rotation angle of the view, in
                                respect to the regular STED angle
        :param spacing:         Voxel size

        :param chunk_size:      A specific chunk size can be defined here in
                                order to optimize the data access, when
                                working with partial images.
        :return:
        """

        assert isinstance(data, numpy.ndarray), "Invalid data format."

        group_name = "psf/" + str(index)
        if group_name not in self.data:
            image_group = self.data.create_group(group_name)
        else:
            image_group = self.data[group_name]

        name = "channel_" + str(channel) + "_scale_" + str(scale)
        if name in image_group:
            return

        if chunk_size is None:
            image_group.create_dataset(name, data=data)
        else:
            image_group.create_dataset(name, data=data, chunks=chunk_size)

        image_group[name].attrs["angle"] = angle
        image_group[name].attrs["spacing"] = spacing
        image_group[name].attrs["size"] = data.shape
        image_group[name].attrs["calculated"] = calculated

    def add_transform(self, scale, index, channel, params, fixed_params, transform_type):
        """
        Adds a spatial transformation as an attribute to the corresponding registered
        view. This means that the registered/resampled image has to be added first, otherwise
        an error is raised.

        Parameters
        ----------
        :param scale            The scale, index and channel parameters identify the
        :param index            registered view, the transform is associated with.
        :param channel

        :param params           The transform parameters. Usually what comes out of
                                transform.GetParameters()
        :param fixed_params     The transform fixed parameters (origin). Usually
                                what comes out of transform.GetFixedParameters()
        :param transform_type
        """
        name = "registered/" + str(index) + "/channel_" + str(channel) + "_scale_" + str(scale)

        if name not in self.data:
            raise ValueError("Dataset %s does not exist" % name)

        self.data[name].attrs["tfm_type"] = transform_type
        self.data[name].attrs["tfm_params"] = params
        self.data[name].attrs["tfm_fixed_params"] = fixed_params

    def add_fused_image(self, data, channel, scale, spacing):
        """
        Add a fused image.

        :param data:            Contains an image from a single rotation angle.
                                The data should be in numpy.ndarray format.
                                Multi-channel data is expected to be a 4D array
                                in which the color channel is the first
                                dimension.
        :param channel          The color channel to be associated with the image.
        :param scale            Percentage from full size. It is possible to save
                                multiple versions of an image in different sizes.
        :param spacing:         Voxel size

        """
        assert isinstance(data, numpy.ndarray), "Invalid data format."

        if "fused" not in self.data:
            image_group = self.data.create_group("fused")
        else:
            image_group = self.data["fused"]

        if channel > 0 and self.channel_count == 1:
            raise ValueError("Invalid channel count")

        name = "channel_" + str(channel) + "_scale_" + str(scale)
        if name in image_group:
            return

        image_group.create_dataset(name, data=data)
        image_group[name].attrs["spacing"] = spacing

    def create_rescaled_images(self, type, scale, chunk_size=None):
        """
        Creates rescaled versions of images of a given type. The scaled images
        are saved directly into the HDF5 file.

        Parameters
        ----------
        type        The image type
        scale       Scale is the percentage of the original image size.
        chunk_size  The same as with the other images. Can be used to define a
                    particular chunk size for data storage.
        """
        if scale in self.get_scales(type):
            if uiutils.get_user_input(
                            "The scale %i already exists for the image type "
                            "%s. Do you want to recalculate?" % (scale, type)
            ):
                pass
            else:
                return
        # Iterate over all the images
        for index in range(self.get_number_of_images(type)):
            group_name = type + "/" + str(index)
            image_group = self.data[group_name]
            # Iterate over channels
            for channel in range(self.channel_count):
                name_new = "channel_" + str(channel) + "_scale_" + str(scale)
                name_ref = "channel_" + str(channel) + "_scale_100"
                # Check if exists and delete if yes.
                if name_new in image_group:
                    del image_group[name_new]
                    continue

                # Zoom
                spacing = tuple(100*x/scale for x in image_group[name_ref].attrs["spacing"])
                z_factor = float(scale)/100
                zoom = (z_factor, ) * self.get_number_of_dimensions()
                data = ndimage.zoom(image_group[name_ref], zoom, order=3)

                if chunk_size is None:
                    image_group.create_dataset(name_new, data=data)
                else:
                    image_group.create_dataset(name_new, data=data, chunks=chunk_size)

                image_group[name_new].attrs["angle"] = image_group[name_ref].attrs["angle"]
                image_group[name_new].attrs["spacing"] = spacing
                image_group[name_new].attrs["size"] = data.shape

    def calculate_missing_psfs(self):
        """
        In case separate PSFs were not recorded for every view, the missing PSFs can be
        calculated here before image fusion. This requires that the spatial transform
        is available for every registered view.
        """
        max_scale = max(self.get_scales("registered"))
        if max_scale < 100:
            if uiutils.get_user_input("There is no registration result "
                                      "available for the original images. The "
                                      "largest available scale is %i. Do you "
                                      "want to proceed with that? " %
                                      max_scale):
                pass
            else:
                raise ValueError("No suitable registration result available.")

        for channel in range(self.channel_count):
            self.set_active_image(0, channel, 100, "psf")
            image_spacing = self.get_voxel_size()
            psf_orig = self.data[self.active_image][:]

            for index in range(1, self.get_number_of_images("registered")):
                if not self.check_if_exists("psf", index, channel, 100):
                    self.set_active_image(index, channel, max_scale, "registered")
                    transform = self.get_transform()
                    psf_new = itkutils.rotate_psf(psf_orig,
                                                  transform,
                                                  image_spacing,
                                                  return_numpy=True)
                    self.add_psf(psf_new, 100, index, channel,
                                 self.get_rotation_angle(), image_spacing,
                                 calculated=True)

    def copy_registration_result(self, from_scale, to_scale):
        """
        With this function it is possible
        to migrate the registration results from one scale to another.

        With very large images it is sometimes easier and faster to perform
        image registration with downsampled versions of the original images.
        The accuracy of the registration result is often very good, even with
        60 percent downsampled images.

        Parameters
        ----------
        :item from_scale    The scale for which there is an existing
                            registration result.
        :item to_scale      The scale for which the new registration results
                            should be calculated.

        Returns
        -------

        """

        # Check that the registration result for the specified scale
        # exists.
        assert from_scale in self.get_scales("registered")
        print("Copying registration results from %i to %i percent scale" % (
              from_scale, to_scale))
        if to_scale not in self.get_scales("original"):
            self.create_rescaled_images("original", to_scale)

        for channel in range(self.channel_count):
            print("Resampling view 0")
            self.set_active_image(0, channel, to_scale, "original")
            self.add_registered_image(self.data[self.active_image][:], to_scale,
                                      0, channel, 0, self.get_voxel_size())
            self.set_active_image(0, channel, to_scale, "registered")
            reference = self.get_itk_image()

            for view in range(1, self.get_number_of_images("original")):
                print("Resampling view %i" % view)
                self.set_active_image(view, channel, from_scale, "registered")
                transform = self.get_transform()
                transform_params = self.get_transform_parameters()
                self.set_active_image(view, channel, to_scale, "original")
                image = self.get_itk_image()
                angle = self.get_rotation_angle(radians=False)
                spacing = self.get_voxel_size()
                result = itkutils.convert_from_itk_image(
                    itkutils.resample_image(image, transform, reference=reference)
                )[0]
                self.add_registered_image(result, to_scale, view, channel, angle,
                                          spacing)
                self.add_transform(to_scale, view, channel, transform_params[0], transform_params[1], transform_params[2])

                #  def add_transform(self, scale, index, channel, params, fixed_params, transform_type):

    def get_rotation_angle(self, radians=True):
        """
        Get rotation angle of the currently active image.

        Parameters
        ----------
        radians     Use radians instead of degrees

        Returns
        -------
        Returns the rotation angle, as degrees or radians
        """
        if radians:
            angle = numpy.pi * int(self.data[self.active_image].attrs["angle"]) / 180
            return angle
        else:
            return int(self.data[self.active_image].attrs["angle"])

    def get_voxel_size(self):
        """
        Get the voxel size of the currently active image.

        Returns
        -------
        Voxel size as a three element tuple (assuming 3D image).
        """
        return list(self.data[self.active_image].attrs["spacing"])

    def get_max(self):
        return self.data[self.active_image][:].max()

    def get_image_size(self):
        """
        Get dimensions of the currently active image

        Returns
        -------
        Image dimensions as a tuple.
        """
        return self.data[self.active_image].attrs["size"]

    def get_dtype(self):
        """
        Get the datatype of the currently acitve image

        Returns
        -------
        The datatype as a numpy.dtype compatible parameter
        """
        return self.data[self.active_image].dtype

    def get_number_of_images(self, image_type):
        """
        Get the number of images of a given type stored in the data structure

        Parameters
        ----------
        :param image_type     The image type

        Returns
        -------
        The number of images of a given type.

        """
        assert image_type in image_types_c
        return len(self.data[image_type])

    def get_number_of_dimensions(self):
        return self.data[self.active_image].ndim

    def get_scales(self, image_type):
        """
        Get a list of the image sizes available for a given image type.

        Parameters
        ----------
        :param image_type       The image type

        Returns
        -------
        Returns a list of the saved scales. Raises an error if the scales
        have been saved inconsistently, i.e. all images of the same type
        do not have the same scales available.
        """

        assert image_type in image_types_c
        scales = []

        def find_scale(name):
            scales.append(int(name.split("_")[-1]))

        for index in range(self.get_number_of_images(image_type)):
            scales = []
            group_name = image_type + "/" + str(index)
            image_group = self.data[group_name]
            image_group.visit(find_scale)

            if index == 0:
                scales_ref = scales
            else:
                if set(scales_ref) != set(scales):
                    raise ValueError("Database error. Resampled images have not been"
                                     "saved consistently for image type %s" %
                                     image_type)

        return scales

    def get_transform(self):
        """
        Get the spatial transformation of the current registered view. Requires a
        registered view to be set as active.

        Returns
        -------
        Return an ITK spatial transform.
        """
        assert "registered" in self.active_image
        tfm_type = self.data[self.active_image].attrs["tfm_type"]
        tfm_params = self.data[self.active_image].attrs["tfm_params"]
        tfm_fixed_params = self.data[self.active_image].attrs["tfm_fixed_params"]
        ndim = self.get_number_of_dimensions()

        return itkutils.make_itk_transform(tfm_type,
                                           ndim,
                                           tfm_params,
                                           tfm_fixed_params)

    def get_transform_parameters(self):
        assert "registered" in self.active_image
        tfm_type = self.data[self.active_image].attrs["tfm_type"]
        tfm_params = self.data[self.active_image].attrs["tfm_params"]
        tfm_fixed_params = self.data[self.active_image].attrs["tfm_fixed_params"]

        return tfm_params, tfm_fixed_params, tfm_type

    def set_active_image(self, index, channel, scale, image_type):
        """
        Select which view is currently active.

        :param index:       View index, goes from 0 to number of views - 1
        :param channel      The currently active color channel. Goes from 0 to
                            number of channels - 1
        :param scale        Image size, as a percentage of the full size.
        :param image_type   Image type as a string, listed in image_types_c
        """
        if int(index) >= self.series_count:
            print("Invalid index. There are only %i images in the file" % self.series_count)
            return
        elif image_type not in image_types_c:
            print("Unkown image type.")
            return
        else:
            if image_type == "fused":
                self.active_image = image_type + "/channel_" + str(channel) + "_scale_" + str(scale)
            else:
                self.active_image = image_type + "/" + str(index) + "/channel_" + str(channel) + "_scale_" + str(scale)
            if self.active_image not in self.data:
                raise ValueError("No such image: %s" % self.active_image)

    # def set_fused_block(self, block, start_index):
    #     assert isinstance(block, numpy.ndarray) and isinstance(start_index, numpy.ndarray)
    #     stop_index = start_index + block.shape
    #     self.data["fused"][start_index:stop_index] = block

    def get_registered_block(self, block_size, block_pad, block_start_index):
        """
        When fusing large images, it is often necessary to divide the images
        into several blocks in order to keep the memory requirements at bay.
        For such use cases functionality was added here to read a partial image
        of a given block size and start index directly from disk. Padding is
        supported as well.

        Parameters
        ----------
        :param block_size   The size of the desired block
        :param block_pad    The amount of padding to be applied to the sides of the
                            block. This kind of partial overlap of adjacent blocks is
                            needed to avoid fusion artifacts at the block boundaries.

        :param block_start_index
                            The index pointing to the beginning of the block.

        Returns
        -------
        The padded image block as a 3D numpy array.

        """
        assert isinstance(block_size, numpy.ndarray)

        assert "registered" in self.active_image, "You must specify a registered image"

        image_size = self.data[self.active_image].shape

        # Apply padding
        end_index = block_start_index + block_size + block_pad
        start_index = block_start_index - block_pad
        # print "Getting a block from ", self.active_image
        # print "The start index is %i %i %i" % tuple(start_index)
        # print "The block size is %i %i %i" % tuple(block_size)

        block_idx = arrayutils.start_to_stop_idx(start_index, end_index)

        if (image_size >= end_index).all() and (start_index >= 0).all():
            block = self.data[self.active_image][block_idx]
            return block

        else:
            pad_block_size = block_size + 2 * block_pad
            block = numpy.zeros(pad_block_size)

            # If the start_index is close to the image boundaries, it is very
            # probable that padding will introduce negative start_index values.
            # In such case the first pixel index must be corrected.
            if (start_index < 0).any():
                block_start = numpy.negative(start_index.clip(max=0))
                image_start = start_index + block_start
            else:
                block_start = (0, 0, 0)
                image_start = start_index
            # If the padded block is larger than the image size the
            # block_size must be adjusted.
            if not (image_size >= end_index).all():
                block_crop = end_index - image_size
                block_crop[block_crop < 0] = 0
                block_end = pad_block_size - block_crop
            else:
                block_end = pad_block_size

            end_index = start_index + block_end

            block_read_idx = arrayutils.start_to_stop_idx(image_start, end_index)
            block_write_idx = arrayutils.start_to_stop_idx(block_start, block_end)

            block[block_write_idx] = self.data[self.active_image][block_read_idx]
            return block

    def get_itk_image(self):
        """
        Get the currently active image as a ITK image instead of a Numpy array.
        """
        return itkutils.convert_from_numpy(self.data[self.active_image][:],

                                           self.data[self.active_image].attrs["spacing"])

    def get_image(self):
        """
        Get the currently active image as an Image object instead of a Numpy array
        """
        return Image(self.data[self.active_image][:],
                     self.data[self.active_image].attrs["spacing"])

    def get_active_image_index(self):
        """
        Get the image of the currently active image
        """
        return self.active_image.split('/')[1]

    def close(self):
        """
        Close the file object.
        """
        self.data.attrs["series_count"] = self.series_count
        self.data.attrs["channel_count"] = self.channel_count

        self.data.close()

    def check_if_exists(self, image_type, index, channel, scale):
        """
        Check if the specified image already exists in the data structure.

        Parameters
        ----------
        :param image_type         The parameters needed to identify an image in the
        :param index        data structure.
        :param channel
        :param scale

        Returns
        -------
        True if Yes, False if No.
        """
        name = image_type + "/" + str(index) + "/channel_" + str(channel) + "_scale_" + str(scale)
        return name in self.data

    def __getitem__(self, item):
        return self.data[self.active_image][item]

    def __setitem__(self, key, value):
        self.data[self.active_image][key] = value


"""
fusion.py

Copyright (C) 2014, 2016 Sami Koho
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license.  See the LICENSE file for details.

This file contains the miplib multi-view image fusion algorithms.
They have been inteded for use with computers that do not support
hardware GPU acceleration. THe accelerated versions of the same functions
can be found in fusion_cuda.py. The fftconvolve function that is
used in this file, can take advantage of MKL optimizations available
in the Anaconda Accelerate package.

"""
import itertools
import os
import shutil
import tempfile
import time

import numpy
import pandas

import miplib.processing.ops_ext as ops_ext
from scipy.ndimage.interpolation import zoom
from scipy.signal import fftconvolve, medfilt

import miplib.processing.ndarray as ops_array
import miplib.processing.to_string as ops_output
from miplib.data.containers import image_data, image
from miplib.data.containers.image import Image
from . import utils as fusion_utils


class MultiViewFusionRL(object):
    """
    The Richardson-Lucy fusion is a result of simultaneous deblurring of
    several 3D volumes.
    """

    def __init__(self, data, writer, options):
        """
        :param data:    a ImageData object

        :param options: command line options that control the behavior
                        of the fusion algorithm
        """
        assert isinstance(data, image_data.ImageData)

        self.data = data
        self.options = options
        self.writer = writer

        # Select views to fuse
        if self.options.fuse_views == -1:
            self.views = range(self.data.get_number_of_images("registered"))
        else:
            self.views = self.options.fuse_views

        self.n_views = len(self.views)

        # Get weights
        self.weights = numpy.zeros(self.n_views, dtype=numpy.float32)

        for idx, view in enumerate(self.views):
            self.data.set_active_image(view, self.options.channel,
                                       self.options.scale, "registered")

            self.weights[idx] = self.data.get_max()

        self.weights /= self.weights.sum()


        # Get image size
        self.data.set_active_image(0, self.options.channel, self.options.scale,
                                   "registered")
        self.image_size = self.data.get_image_size()
        self.imdims = len(self.image_size)

        print("The original image size is {}".format(tuple(self.image_size)))

        self.voxel_size = self.data.get_voxel_size()
        self.iteration_count = 0

        # Setup blocks
        self.num_blocks = options.num_blocks
        self.block_size, self.image_size = self.__calculate_block_and_image_size()
        self.memmap_directory = tempfile.mkdtemp()

        # Memmap the estimates to reduce memory requirements. This will slow
        # down the fusion process considerably..
        if self.options.memmap_estimates:
            estimate_new_f = os.path.join(self.memmap_directory, "estimate_new.dat")
            self.estimate_new = Image(numpy.memmap(estimate_new_f, dtype='float32',
                                                   mode='w+',
                                                   shape=tuple(self.image_size)), self.voxel_size)

            estimate_f = os.path.join(self.memmap_directory, "estimate.dat")
            self.estimate = Image(numpy.memmap(estimate_f, dtype=numpy.float32,
                                               mode='w+',
                                               shape=tuple(self.image_size)), self.voxel_size)
        else:
            self.estimate = Image(numpy.zeros(tuple(self.image_size),
                                              dtype=numpy.float32), self.voxel_size)
            self.estimate_new = Image(numpy.zeros(tuple(self.image_size),
                                                  dtype=numpy.float32), self.voxel_size)

        if not self.options.disable_tau1:
            prev_estimate_f = os.path.join(self.memmap_directory, "prev_estimate.dat")
            self.prev_estimate = numpy.memmap(prev_estimate_f, dtype=numpy.float32,
                                              mode='w+',
                                              shape=tuple(self.image_size))
        # Setup PSFs
        self.psfs = []
        self.adj_psfs = []
        self.__get_psfs()
        if "opt" in self.options.fusion_method:
            self.virtual_psfs = []
            self.__compute_virtual_psfs()
        else:
            pass

        print("The fusion will be run with %i blocks" % self.num_blocks)
        padded_block_size = tuple(i + 2 * self.options.block_pad for i in self.block_size)
        print("The internal block size is %s" % (padded_block_size,))

        self.column_headers = ('t', 'tau1', 'leak', 'e',
                               's', 'u', 'n', 'uesu')
        self._progress_parameters = numpy.empty((self.options.max_nof_iterations, len(self.column_headers)),
                                                dtype=numpy.float32)

    @property
    def progress_parameters(self):
        return pandas.DataFrame(data=self._progress_parameters, columns=self.column_headers)

    def compute_estimate(self):
        """
        Calculates a single RL fusion estimate. There is no reason to call this
        function -- it is used internally by the class during fusion process.
        """

        print('Beginning the computation of the %i. estimate' % self.iteration_count)

        if "multiplicative" in self.options.fusion_method:
            self.estimate_new[:] = numpy.float32(1.0)
        else:
            self.estimate_new[:] = numpy.float32(0)

        # Iterate over views
        for idx, view in enumerate(self.views):

            # Get PSFs for view
            psf = self.psfs[idx]
            adj_psf = self.adj_psfs[idx]

            self.data.set_active_image(view, self.options.channel,
                                       self.options.scale, "registered")

            #weighting = float(self.data.get_max()) / 255
            weighting = self.weights[idx]

            iterables = (range(0, m, n) for m, n in zip(self.image_size, self.block_size))
            pad = self.options.block_pad
            cache_idx = tuple(slice(pad, pad + block) for block in self.block_size)

            # Iterate over blocks
            for pos in itertools.product(*iterables):

                estimate_idx = tuple(slice(j, j + k) for j, k in zip(pos, self.block_size))
                index = numpy.array(pos, dtype=int)
                if self.options.block_pad > 0:
                    estimate_block = self.get_padded_block(index.copy())
                else:
                    estimate_block = self.estimate[estimate_idx]

                # Execute: cache = convolve(PSF, estimate), non-normalized
                estimate_block_new = fftconvolve(estimate_block, psf, mode='same')
                estimate_block_new *= weighting

                # Execute: cache = data/cache
                block = self.data.get_registered_block(self.block_size,
                                                       self.options.block_pad,
                                                       index.copy())

                with numpy.errstate(divide="ignore"):
                    estimate_block_new = block / estimate_block_new
                    estimate_block_new[estimate_block_new == numpy.inf] = 0.0
                    estimate_block_new = numpy.nan_to_num(estimate_block_new)

                # Execute: cache = convolve(PSF(-), cache), inverse of non-normalized
                # Convolution with virtual PSFs is performed here as well, if
                # necessary
                estimate_block_new = fftconvolve(estimate_block_new, adj_psf, mode='same')

                # Update the contribution from a single view to the new estimate
                if self.options.block_pad == 0:
                    if "multiplicative" in self.options.fusion_method:
                        self.estimate_new[estimate_idx] *= estimate_block_new
                    else:
                        self.estimate_new[estimate_idx] += estimate_block_new
                else:
                    if "multiplicative" in self.options.fusion_method:
                        self.estimate_new[estimate_idx] *= estimate_block_new[cache_idx]

                    else:
                        # print "The block size is ", self.block_size
                        self.estimate_new[estimate_idx] += estimate_block_new[cache_idx]

        # I changed the weighting scheme a little bit
        # I'm not sure if this thing is necessary; maybe in the multiplicative?
        # Divide with the number of projections
        # if "summative" in self.options.fusion_method:
        #     self.estimate_new *= (1.0 / self.n_views)
        # else:
        #     self.estimate_new[:] = ops_array.nroot(self.estimate_new,
        #                                            self.n_views)

        return ops_ext.update_estimate_poisson(self.estimate,
                                               self.estimate_new,
                                               self.options.convergence_epsilon)

    def execute(self):
        """
        This is the main fusion function
        """

        print("Preparing image fusion.")

        save_intermediate_results = self.options.save_intermediate_results

        first_estimate = self.options.first_estimate

        self.data.set_active_image(0,
                                   self.options.channel,
                                   self.options.scale,
                                   "registered")

        if first_estimate == 'first_image':
            self.estimate[:] = self.data[:].astype(numpy.float32)
        elif first_estimate == 'first_image_mean':
            self.estimate[:] = numpy.float32(numpy.mean(self.data[:]))
        elif first_estimate == 'sum_of_originals':
            self.estimate[:] = fusion_utils.sum_of_all(self.data,
                                                       self.options.channel,
                                                       self.options.scale)
        elif first_estimate == 'sum_of_registered':
            self.estimate[:] = fusion_utils.sum_of_all(self.data,
                                                       self.options.channel,
                                                       self.options.scale,
                                                       "registered")
        elif first_estimate == 'simple_fusion':
            self.estimate[:] = fusion_utils.simple_fusion(self.data,
                                                          self.options.channel,
                                                          self.options.scale)
        elif first_estimate == 'average_af_all':
            self.estimate[:] = fusion_utils.average_of_all(self.data,
                                                           self.options.channel,
                                                           self.options.scale,
                                                           "registered")
        elif first_estimate == 'constant':
            self.estimate[:] = numpy.float32(self.options.estimate_constant)
        else:
            raise NotImplementedError(repr(first_estimate))

        self.iteration_count = 0
        max_count = self.options.max_nof_iterations
        initial_photon_count = self.data[:].sum()

        bar = ops_output.ProgressBar(0,
                                     max_count,
                                     totalWidth=40,
                                     show_percentage=False)

        self._progress_parameters = numpy.zeros((self.options.max_nof_iterations,
                                                 len(self.column_headers)),
                                                dtype=numpy.float32)

        # The Fusion calculation starts here
        # ====================================================================
        try:
            while True:

                info_map = {}
                ittime = time.time()

                if not self.options.disable_tau1:
                    self.prev_estimate[:] = self.estimate.copy()

                e, s, u, n = self.compute_estimate()

                self.iteration_count += 1
                photon_leak = 1.0 - (e + s + u) / initial_photon_count
                u_esu = u / (e + s + u)

                if not self.options.disable_tau1:
                    tau1 = abs(self.estimate - self.prev_estimate).sum() / abs(
                        self.prev_estimate).sum()
                    info_map['TAU1=%s'] = tau1

                t = time.time() - ittime
                leak = 100 * photon_leak

                # Update UI
                info_map['E/S/U/N=%s/%s/%s/%s'] = int(e), int(s), int(u), int(n)
                info_map['LEAK=%s%%'] = leak
                info_map['U/ESU=%s'] = u_esu
                info_map['TIME=%ss'] = t
                bar.updateComment(' ' + ', '.join([k % (ops_output.tostr(info_map[k])) for k in sorted(info_map)]))
                bar(self.iteration_count)
                print()

                # Save parameters to file
                self._progress_parameters[self.iteration_count - 1] = (t, tau1, leak, e, s, u, n, u_esu)

                # Save intermediate image
                if save_intermediate_results:
                    self.writer.write(Image(self.estimate, self.voxel_size))

                # Check if it's time to stop:
                if int(u) == 0 and int(n) == 0:
                    stop_message = 'The number of non converging photons reached to zero.'
                    break
                elif self.iteration_count >= max_count:
                    stop_message = 'The number of iterations reached to maximal count: %s' % max_count
                    break
                elif not self.options.disable_tau1 and tau1 <= self.options.rltv_stop_tau:
                    stop_message = 'Desired tau-threshold achieved'
                    break
                else:
                    continue

        except KeyboardInterrupt:
            stop_message = 'Iteration was interrupted by user.'

        # if self.num_blocks > 1:
        #     self.estimate = self.estimate[0:real_size[0], 0:real_size[1], 0:real_size[2]]

        print()
        bar.updateComment(' ' + stop_message)
        bar(self.iteration_count)
        print()

    # region Prepare PSFs
    def __get_psfs(self):
        """
        Reads the PSFs from the HDF5 data structure and zooms to the same pixel
        size with the registered images, of selected scale and channel.
        """
        self.data.set_active_image(0, self.options.channel,
                                   self.options.scale, "registered")
        image_spacing = self.data.get_voxel_size()

        for i in self.views:
            self.data.set_active_image(i, 0, 100, "psf")
            psf_orig = self.data[:]
            psf_spacing = self.data.get_voxel_size()

            # Zoom to the same voxel size
            zoom_factors = tuple(x / y for x, y in zip(psf_spacing, image_spacing))
            psf_new = zoom(psf_orig, zoom_factors).astype(numpy.float32)

            psf_new /= psf_new.sum()

            # Save the zoomed and rotated PSF, as well as its mirrored version
            self.psfs.append(psf_new)

            if self.imdims == 3:
                self.adj_psfs.append(psf_new[::-1, ::-1, ::-1])
            else:
                self.adj_psfs.append(psf_new[::-1, ::-1])

    def __compute_virtual_psfs(self):
        """
        Implements a Virtual PSF calculation routine, as described in "Efficient
        Bayesian-based multiview deconvolution" by Preibich et al in Nature
        Methods 11/6 (2014)
        """

        print("Caclulating Virtual PSFs")

        for i in range(self.n_views):
            virtual_psf = numpy.ones(self.psfs[0].shape, dtype=self.psfs[0].dtype)
            for j in range(self.n_views):
                if j == i:
                    pass
                else:
                    cache = fftconvolve(
                        fftconvolve(
                            self.adj_psfs[i],
                            self.psfs[j],
                            mode='same'
                        ),
                        self.adj_psfs[j],
                        mode='same'

                    )

                    virtual_psf *= cache.real

            virtual_psf *= self.adj_psfs[i]
            virtual_psf /= virtual_psf.sum()
            self.adj_psfs[i] = virtual_psf
            # self.virtual_psfs.append(virtual_psf)

    # endregion


    def __calculate_block_and_image_size(self):
        """
        Calculate the block size and the internal image size for a given
        number of blocks. 1,2,4 or 8 blocks are currently supported.

        """
        block_size = self.image_size
        image_size = self.image_size

        if self.num_blocks == 1:
            return block_size, image_size
        elif self.num_blocks == 2:
            multiplier3 = numpy.array([2, 1, 1])
            multiplier2 = numpy.array([2, 1])
        elif self.num_blocks == 4:
            multiplier3 = numpy.array([4, 1, 1])
            multiplier2 = numpy.array([2, 2])
        elif self.num_blocks == 8:
            multiplier3 = numpy.array([4, 2, 1])
            multiplier2 = numpy.array([4, 2])
        elif self.num_blocks == 12:
            multiplier3 = numpy.array([4, 2, 2])
            multiplier2 = numpy.array([4, 3])
        elif self.num_blocks == 24:
            multiplier3 = numpy.array([4, 3, 2])
            multiplier2 = numpy.array([6, 4])
        elif self.num_blocks == 48:
            multiplier3 = numpy.array([4, 4, 3])
            multiplier2 = numpy.array([8, 6])
        elif self.num_blocks == 64:
            multiplier3 = numpy.array([4, 4, 4])
            multiplier2 = numpy.array([8, 8])
        elif self.num_blocks == 96:
            multiplier3 = numpy.array([6, 4, 4])
            multiplier2 = numpy.array([12, 8])
        elif self.num_blocks == 144:
            multiplier3 = numpy.array([4, 6, 6])
            multiplier2 = numpy.array([12, 12])
        else:
            raise NotImplementedError

        if self.imdims == 2:
            block_size = numpy.ceil(self.image_size.astype(numpy.float16) / multiplier2).astype(numpy.int64)
            image_size += (multiplier2 * block_size - image_size)
        else:
            block_size = numpy.ceil(self.image_size.astype(numpy.float16) / multiplier3).astype(numpy.int64)
            image_size += (multiplier3 * block_size - image_size)

        return block_size, image_size

    def get_padded_block(self, image, block_start_index):
        """
        Get a padded block from the self.estimate

        Parameters
        ----------
        :param image: a numpy.ndarray or or its subclass
        :param block_start_index  The real block start index, not considering the padding

        Returns
        -------
        Returns the padded estimate block as a numpy array.

        """

        block_pad = self.options.block_pad
        image_size = self.image_size
        ndims = self.imdims

        # Apply padding
        end_index = block_start_index + self.block_size + block_pad
        start_index = block_start_index - block_pad

        idx = tuple(slice(start, stop) for start, stop in zip(start_index, end_index))

        # If the padded block fits within the image boundaries, nothing special
        # is needed to extract it. Normal numpy slicing notation is used.
        if (image_size >= end_index).all() and (start_index >= 0).all():
            return image[idx]

        else:
            block_size = tuple(i + 2 * block_pad for i in self.block_size)
            # Block outside the image boundaries will be filled with zeros.
            block = numpy.zeros(block_size)
            # If the start_index is close to the image boundaries, it is very
            # probable that padding will introduce negative start_index values.
            # In such case the first pixel index must be corrected.
            if (start_index < 0).any():
                block_start = numpy.negative(start_index.clip(max=0))
                image_start = start_index + block_start
            else:
                block_start = (0,) * ndims
                image_start = start_index

            # If the padded block is larger than the image size the
            # block_size must be adjusted.
            if not (image_size >= end_index).all():
                block_crop = end_index - image_size
                block_crop[block_crop < 0] = 0
                block_end = block_size - block_crop
            else:
                block_end = block_size

            end_index = start_index + block_end

            block_idx = tuple(slice(start, stop) for start, stop in zip(block_start, block_end))
            image_idx = tuple(slice(start, stop) for start, stop in zip(image_start, end_index))

            block[block_idx] = image[image_idx]

            return block

    # region Get Output
    def get_result(self, cast_to_8bit=False):
        """
        Show fusion result. This is a temporary solution for now
        calling Fiji through ITK. An internal viewer would be
        preferable.
        """
        if cast_to_8bit:
            result = self.estimate.copy()
            result *= (255.0 / result.max())
            result[result < 0] = 0
            return Image(result, self.voxel_size)

        return Image(self.estimate, self.voxel_size)

    def save_to_hdf(self):
        """
        Save result to the miplib data structure.

        """
        self.data.set_active_image(0,
                                   self.options.channel,
                                   self.options.scale,
                                   "registered")
        spacing = self.data.get_voxel_size()

        self.data.add_fused_image(self.estimate,
                                  self.options.channel,
                                  self.options.scale,
                                  spacing)

    # endregion

    def close(self):
        if self.options.memmap_estimates:
            del self.estimate
            del self.estimate_new
        if not self.options.disable_tau1:
            del self.prev_estimate

        shutil.rmtree(self.memmap_directory)



class FixedDictionary(object):
    """
    A dictionary with immutable keys. Is initialized at construction
    with a list of key values.
    """
    def __init__(self, keys):
        assert isinstance(keys, list) or isinstance(keys, tuple)
        self._dictionary = dict.fromkeys(keys)

    def __setitem__(self, key, value):
        if key not in self._dictionary:
            raise KeyError("The key {} is not defined".format(key))
        else:
            self._dictionary[key] = value

    def __getitem__(self, key):
        return self._dictionary[key]

    @property
    def keys(self):
        return list(self._dictionary.keys())

    @property
    def contents(self):
        return list(self._dictionary.keys()), list(self._dictionary.values())



itk_transforms_c = {
    'sitkIdentity'          :   0,
    'sitkTranslation'       :   1,
    'sitkScale'             :   2,
    'sitkScaleLogarithmic'  :   3,
    'sitkEuler'             :   4,
    'sitkSimilarity'        :   5,
    'sitkQuaternionRigid'   :   6,
    'sitkVersor'            :   7,
    'sitkVersorRigid'       :   8,
    'sitkScaleSkewVersor'   :   9,
    'sitkAffine'            :   10,
    'sitkComposite'         :   11,
    'sitkDisplacementField' :   12,
    'sitkBSplineTransform'  :   13
}

image_types_c = ("original", "registered", "fused", "psf")
params_c = ("angle", "scale", "index", "channel")



import argparse
from miplib.psf import psfgen
import miplib.ui.cli.argparse_helpers as helpers

def parse_psf_type(args):
    if args == "confocal" or args == "sted":
        return psfgen.GAUSSIAN | psfgen.CONFOCAL
    elif args == "widefield":
        return psfgen.GAUSSIAN | psfgen.WIDEFIELD
    else:
        raise argparse.ArgumentTypeError("Unknown PSF type")


def get_psf_estimation_options_group(parser):
    assert isinstance(parser, argparse.ArgumentParser)
    group = parser.add_argument_group(
        "PSF estimation", 
        "Options for controlling the PSF estimation algorithm")

    group.add_argument(
        '--psf-type',
        type=parse_psf_type,
        default=psfgen.GAUSSIAN | psfgen.CONFOCAL
    )

    group.add_argument(
        '--psf-shape',
        type=helpers.parse_int_tuple,
        default=(256,256)

    )
    group.add_argument(
        '--psf-size',
        type=helpers.parse_float_tuple,
        default=(4., 4.)
    )
    group.add_argument(
        '--ex-wl',
        type=float,
        default=488
    )
    group.add_argument(
        '--em-wl',
        type=float,
        default=550
    )
    group.add_argument(
        '--na',
        type=float,
        default=1.4
    )
    group.add_argument(
        '--refractive-index',
        type=float,
        default=1.414
    )
    group.add_argument(
        '--magnification',
        type=float,
        default=1.0
    )
    group.add_argument(
        '--pinhole-radius',
        type=float,
        default=None

    )

    return parser



"""
fuse.py

Copyright (c) 2016 Sami Koho  All rights reserved.

This software may be modified and distributed under the terms
of the BSD license.  See the LICENSE file for details.

This is the main program file for the miplib fusion calculation
"""

import os
import sys
import time

import miplib.data.containers.image_data as image_data
import miplib.processing.fusion.fusion as fusion
import miplib.processing.fusion.fusion_cuda as gpufusion
import miplib.processing.to_string as genutils
import miplib.ui.cli.miplib_entry_point_options as arguments
import miplib.ui.utils as uiutils


def main():

    options = arguments.get_fusion_script_options(sys.argv[1:])
    full_path = os.path.join(options.working_directory,
                             options.data_file)

    if not os.path.isfile(full_path):
        raise AttributeError("No such file: %s" % full_path)
    elif not full_path.endswith(".hdf5"):
        raise AttributeError("Not a HDF5 file")

    data = image_data.ImageData(full_path)

    if options.scale not in data.get_scales("registered"):
        print("Images at the defined scale do not exist in the data structure." \
              "The original images will be now resampled. This may take a long" \
              "time depending on the image size and the number of views.")
        data.create_rescaled_images("registered", options.scale)

    if data.get_number_of_images("psf") != data.get_number_of_images("registered"):
        print("Some PSFs are missing. They are going to be calculated from the " \
              "original STED PSF (that is assumed to be at index 0).")
        data.calculate_missing_psfs()

    if not options.disable_cuda:
        print("Trying to run the image fusion with GPU acceleration.")
        task = gpufusion.MultiViewFusionRLCuda(data, options)
    else:
        task = fusion.MultiViewFusionRL(data, options)

    # task = fusion.MultiViewFusionRL(data, options)
    begin = time.time()
    task.execute()
    end = time.time()

    if options.evaluate_results:
        task.show_result()

    print("Fusion complete.")
    print("The fusion process with %i iterations " \
          "took %s (H:M:S) to complete." % (options.max_nof_iterations,
                                            genutils.format_time_string(
                                                end-begin)))

    if uiutils.get_user_input("Do you want to save the result to TIFF? "):
        file_path = os.path.join(options.working_directory,
                                 "fusion_result.tif")
        task.save_to_tiff(file_path)

    if uiutils.get_user_input("Do you want to save the result to the HDF data "
                              "structure? "):
        task.save_to_hdf()

    task.close()
    data.close()


if __name__ == "__main__":
    main()



# coding=utf-8
"""
Sami Koho 01/2018

Sectioned Fourier Shell Correlation for complex resolution analysis
in 3D images.
"""
from scipy.ndimage.interpolation import rotate
import numpy as np
import miplib.data.iterators.fourier_shell_iterators as iterators

import miplib.data.containers.fourier_correlation_data as containers
import miplib.processing.ndarray as ndarray
import miplib.processing.image as imops
from miplib.data.containers.image import Image
from miplib.processing import windowing
from . import analysis as fsc_analysis
from math import floor


def calculate_fourier_plane_correlation(image1, image2, args, z_correction=1):
        steps = np.arange(0, 360, args.d_angle)
        data = containers.FourierCorrelationDataCollection()

        for idx, step in enumerate(steps):
            im1_rot = np.fft.fftshift(np.fft.fftn(rotate(image1, step, reshape=False)))
            im2_rot = np.fft.fftshift(np.fft.fftn(rotate(image2, step, reshape=False)))

            numerator = np.sum(im1_rot*np.conjugate(im2_rot), axis=(0,2))
            denominator = np.sum(np.sqrt(np.abs(im1_rot)**2 * np.abs(im2_rot)**2), axis=(0,2))

            correlation = ndarray.safe_divide(numerator, denominator)

            zero = correlation.size / 2
            correlation = correlation[zero:]

            result = containers.FourierCorrelationData()
            result.correlation["correlation"] = correlation
            result.correlation["frequency"] = np.linspace(0, 1.0, num=correlation.size)
            result.correlation["points-x-bin"] = np.ones(correlation.size)*(im2_rot.shape[2]*im2_rot.shape[0])

            data[int(step)] = result

        analyzer = fsc_analysis.FourierCorrelationAnalysis(data, image1.spacing[0], args)
        return analyzer.execute(z_correction=z_correction)



def calculate_one_image_sectioned_fsc(image, args, z_correction=1):
    """ A function to calculate one-image sectioned FSC. I assume here that prior to calling the function,
    the image is going to be in a correct shape, resampled to isotropic spacing and zero padded. If the image
    dimensions are wrong (not a cube) the function will return an error.
    
    :param image: a 3D image, with isotropic spacing and cubic shape
    :type image: Image
    :param options: options for the FSC calculation
    :type options: argparse options
    :param z_correction: correction, for anisotropic sampling. It is the ratio of axial vs. lateral spacing, defaults to 1
    :type z_correction: float, optional
    :return: the resolution measurement results organized by rotation angle
    :rtype: FourierCorrelationDataCollection object
    """
    assert isinstance(image, Image)
    assert all(s == image.shape[0] for s in image.shape)
    
    image1, image2 = imops.checkerboard_split(image)

    image1 = Image(windowing.apply_hamming_window(image1), image1.spacing)
    image2 = Image(windowing.apply_hamming_window(image2), image2.spacing)
    
    iterator = iterators.AxialExcludeHollowConicalFourierShellIterator(image1.shape, args.d_bin, 
                                                                      args.d_angle, args.d_extract_angle)
    fsc_task = DirectionalFSC(image1, image2, iterator)

    data = fsc_task.execute()
    
    analyzer = fsc_analysis.FourierCorrelationAnalysis(data, image1.spacing[0], args)
    result = analyzer.execute(z_correction=z_correction)
    
    def func(x, a, b, c, d):
        return a * np.exp(c * (x - b)) + d

    params = [0.95988146, 0.97979108, 13.90441896, 0.55146136]

    for angle, dataset in result:
        point = dataset.resolution["resolution-point"][1]

        cut_off_correction = func(point, *params)
        dataset.resolution["spacing"] /= cut_off_correction
        dataset.resolution["resolution"] /= cut_off_correction
    
    return result

def calculate_two_image_sectioned_fsc(image1, image2, args, z_correction=1):
    assert isinstance(image1, Image)
    assert isinstance(image2, Image)

    image1 = Image(windowing.apply_hamming_window(image1), image1.spacing)
    image2 = Image(windowing.apply_hamming_window(image2), image2.spacing)

    iterator = iterators.AxialExcludeHollowConicalFourierShellIterator(image1.shape, args.d_bin, args.d_angle,
                                                                       args.d_extract_angle)
    fsc_task = DirectionalFSC(image1, image2, iterator)
    data = fsc_task.execute()

    analyzer = fsc_analysis.FourierCorrelationAnalysis(data, image1.spacing[0], args)
    return analyzer.execute(z_correction=z_correction)




class DirectionalFSC(object):
    def __init__(self, image1, image2, iterator, normalize_power=False):
        assert isinstance(image1, Image)
        assert isinstance(image2, Image)

        if image1.ndim != 3 or image1.shape[0] <= 1:
            raise ValueError("You should provide a stack for FSC analysis")

        if image1.shape != image2.shape:
            raise ValueError("Image dimensions do not match")

        # Create an Iterator
        self.iterator = iterator

        # FFT transforms of the input images
        self.fft_image1 = np.fft.fftshift(np.fft.fftn(image1))
        self.fft_image2 = np.fft.fftshift(np.fft.fftn(image2))

        if normalize_power:
            pixels = image1.shape[0]**3
            self.fft_image1 /= (np.array(pixels * np.mean(image1)))
            self.fft_image2 /= (np.array(pixels * np.mean(image2)))

        self._result = None

        self.pixel_size = image1.spacing[0]

    @property
    def result(self):
        if self._result is None:
            return self.execute()
        else:
            return self._result

    def execute(self):
        """
        Calculate the FRC
        :return: Returns the FRC results. They are also saved inside the class.
                 The return value is just for convenience.
        """

        data_structure = containers.FourierCorrelationDataCollection()
        radii, angles = self.iterator.steps
        freq_nyq = self.iterator.nyquist
        shape = (angles.shape[0], radii.shape[0])
        c1 = np.zeros(shape, dtype=np.float32)
        c2 = np.zeros(shape, dtype=np.float32)
        c3 = np.zeros(shape, dtype=np.float32)
        points = np.zeros(shape, dtype=np.float32)

        # Iterate through the sphere and calculate initial values
        for ind_ring, shell_idx, rotation_idx in self.iterator:
            subset1 = self.fft_image1[ind_ring]
            subset2 = self.fft_image2[ind_ring]

            c1[rotation_idx, shell_idx] = np.sum(subset1 * np.conjugate(subset2)).real
            c2[rotation_idx, shell_idx] = np.sum(np.abs(subset1) ** 2)
            c3[rotation_idx, shell_idx] = np.sum(np.abs(subset2) ** 2)

            points[rotation_idx, shell_idx] = len(subset1)

        # Finish up FRC calculation for every rotation angle and sav
        # results to the data structure.
        for i in range(angles.size):

            # Calculate FRC for every orientation
            spatial_freq = radii.astype(np.float32) / freq_nyq
            n_points = np.array(points[i])
            frc = ndarray.safe_divide(c1[i], np.sqrt(c2[i] * c3[i]))

            result = containers.FourierCorrelationData()
            result.correlation["correlation"] = frc
            result.correlation["frequency"] = spatial_freq
            result.correlation["points-x-bin"] = n_points

            # Save result to the structure and move to next
            # angle
            data_structure[angles[i]] = result

        return data_structure



#!/usr/bin/env python
# -*- python -*-

"""
register.py

Copyright (c) 2016 Sami Koho  All rights reserved.

This software may be modified and distributed under the terms
of the BSD license.  See the LICENSE file for details.

This is the main program file for the miplib fusion calculation
"""

import os
import sys

import SimpleITK as sitk

from miplib.data.containers import image_data
from miplib.processing import itk as itkutils
from miplib.processing.registration import registration_mv
from miplib.ui import utils
from miplib.ui.cli import miplib_entry_point_options


def main():
    options = miplib_entry_point_options.get_register_script_options(sys.argv[1:])

    # Get Data
    filename = sys.argv[1]
    if not filename.endswith(".hdf5"):
        raise ValueError("Please specify a HDF5 data file")
    full_path = os.path.join(options.working_directory, filename)
    if not os.path.exists(full_path):
        raise ValueError("The specified file %s does not exist" % full_path)

    data = image_data.ImageData(full_path)

    # Check that requested image size exists. If not, create it.
    if options.scale not in data.get_scales("original"):
        print("Images at the defined scale do not exist in the data " \
              "structure. The original images will be now resampled. " \
              "This may take a long time depending on the image size " \
              "and the number of views.")
        data.create_rescaled_images("original", options.scale)

    data.set_active_image(0, options.channel, options.scale, "original")
    spacing = data.get_voxel_size()
    data.add_registered_image(data[:], options.scale, 0, options.channel,
                              0, spacing)

    if options.evaluate_results:
        fixed_image = data.get_itk_image()

    # Setup registration. View number 0 is always the reference for now.
    # The behavior can be easily changed if necessary.
    task = registration_mv.RotatedMultiViewRegistration(data, options)
    task.set_fixed_image(0)

    # Iterate over the rotated views
    for view in range(1, data.get_number_of_images("original")):
        task.set_moving_image(view)
        if data.check_if_exists("registered", view, options.channel,
                                options.scale):
            if utils.get_user_input("There is a saved registration result for "
                                    "the view %i. Do you want to skip it?" %
                                    view):
                continue

        task.execute()

        if options.evaluate_results:
            moving_image = task.get_resampled_result()
            sitk.Show(
                itkutils.make_composite_rgb_image(fixed_image, moving_image))
            if utils.get_user_input(
                    "Do you want to save the result (yes/no)? "):
                task.save_result()
                continue
            else:
                if utils.get_user_input(
                        "Skipping view %i. Do you want to continue "
                        "registration?" % view):
                    continue
                else:
                    print("Exiting registration without saving results")
                    break
        else:
            task.save_result()
            continue

    data.close()


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt


def xy_scatter_plot_with_labels(x, y, labels, size=(3,3)):

    assert len(x) == len(y) == len(labels)

    fig, ax = plt.subplots(figsize=size)
    ax.scatter(x, y)

    for i, txt in enumerate(labels):
        ax.annotate(txt, (x[i], y[i]))

    return fig



import numpy as np

from miplib.data.containers.image_data import ImageData
from miplib.data.containers.image import Image


def sum_of_all(data_structure, channel=0, scale=100, image_type="original"):
    assert isinstance(data_structure, ImageData)

    n_views = data_structure.get_number_of_images(image_type)
    data_structure.set_active_image(0, channel, scale, image_type)
    result = np.zeros(data_structure.get_image_size(), dtype=np.float32)
    pixel_size = data_structure.get_voxel_size()

    for i in range(n_views):
        data_structure.set_active_image(i, channel, scale, image_type)
        result += data_structure[:]

    return Image(result, pixel_size)


def average_of_all(data_structure, channel=0, scale=100, image_type="original"):
    assert isinstance(data_structure, ImageData)
    n_views = data_structure.get_number_of_images(image_type)
    data_structure.set_active_image(0, channel, scale, image_type)
    pixel_size = data_structure.get_voxel_size()

    result = sum_of_all(data_structure, channel, scale, image_type)

    return Image(result/n_views, pixel_size)


def simple_fusion(data_structure, channel=0, scale=100):
    assert isinstance(data_structure, ImageData)
    image_type = "registered"

    n_views = data_structure.get_number_of_images(image_type)
    data_structure.set_active_image(0, channel, scale, image_type)
    pixel_size = data_structure.get_voxel_size()

    result = data_structure[:]

    for i in range(1, n_views):
        data_structure.set_active_image(i, channel, scale, image_type)
        result = (result - (result - data_structure[:]).clip(min=0)).clip(min=0).astype(np.float32)

    return Image(result, pixel_size)


def common_start(sa, sb):
    """ returns the longest common substring from the beginning of sa and sb """
    def _iter():
        for a, b in zip(sa, sb):
            if a == b:
                yield a
            else:
                return

    return ''.join(_iter())

def common_string(strings):
    """
    Find the longest common string.

    :param strings:
    :return:
    """
    prefix1 = strings[0]
    prefix2 = strings[1]

    if prefix1.find('/') != -1:
        prefix1 = prefix1.split('/')
        prefix1 = prefix1[len(prefix1) - 1]

    if prefix2.find('/') != -1:
        prefix2 = prefix2.split('/')
        prefix2 = prefix2[len(prefix2) - 1]

    strings = [prefix1, prefix2]
    prefix = prefix1

    for s in strings:
        if len(s) < len(prefix):
            prefix = prefix[:len(s)]
        if not prefix:
            return ''
        for i in range(len(prefix)):
            if prefix[i] != s[i]:
                prefix = prefix[:i]
                break

    return prefix




import numpy as np


def _nd_window(data, filter_function):
    """
    Performs on N-dimensional spatial-domain data.
    This is done to mitigate boundary effects in the FFT.

    Parameters
    ----------
    data : ndarray
           Input data to be windowed, modified in place.
    filter_function : 1D window generation function
           Function should accept one argument: the window length.
           Example: scipy.signal.hamming
    """
    result = data.copy().astype(np.float64)
    for axis, axis_size in enumerate(data.shape):
        # set up shape for numpy broadcasting
        filter_shape = [1, ] * data.ndim
        filter_shape[axis] = axis_size
        window = filter_function(axis_size).reshape(filter_shape)
        # scale the window intensities to maintain image intensity
        np.power(window, (1.0/data.ndim), out=window)
        result *= window
    return result


def apply_hamming_window(data):
    assert issubclass(data.__class__, np.ndarray)

    return _nd_window(data, np.hamming)
